import asyncio
import logging
import threading
import time
import traceback

from pyroute2 import config

from . import schema
from .events import (
    DBMExitException,
    InvalidateHandlerException,
    RescheduleException,
    ShutdownException,
    State,
)
from .messages import cmsg_event, cmsg_failed

log = logging.getLogger(__name__)


class NDBConfig(dict):
    def __init__(self, task_manager):
        self.task_manager = task_manager

    def __getitem__(self, key):
        return self.task_manager.config_get(key)

    def __setitem__(self, key, value):
        return self.task_manager.config_set(key, value)

    def __delitem__(self, key):
        return self.task_manager.config_del(key)

    def keys(self):
        return self.task_manager.config_keys()

    def items(self):
        return self.task_manager.config_items()

    def values(self):
        return self.task_manager.config_values()


class TaskAdapter:
    def __init__(self):
        self.event = asyncio.Event()
        self.state = State()


class Task:
    def __init__(self, coro, target_state, obj):
        self.exception = None
        self.coro = coro
        self.target_state = target_state
        self.obj = obj if obj is not None else TaskAdapter()
        self.restart()

    def restart(self):
        self.task = asyncio.create_task(self.coro())
        self.obj.task_id = id(self.task)

    def commit(self):
        self.exception = self.task.exception()

    @property
    def event(self):
        return self.obj.event

    @property
    def state(self):
        return self.obj.state

    @property
    def task_id(self):
        return id(self.task)


class TaskManager:
    def __init__(self, ndb):
        self.ndb = ndb
        self.log = ndb.log
        self.event_map = {}
        self.task_map = {}
        self.event_queue = (
            asyncio.Queue()
        )  # LoggingQueue(log=self.ndb.log.channel('queue'))
        self.stop_event = asyncio.Event()
        self.reload_event = asyncio.Event()
        self.thread = None
        self.ctime = self.gctime = time.time()
        self.ready = asyncio.Event()

    def register_handler(self, event, handler):
        if event not in self.event_map:
            self.event_map[event] = []
        self.event_map[event].append(handler)

    def unregister_handler(self, event, handler):
        self.event_map[event].remove(handler)

    async def handler_default(self, sources, target, event):
        if isinstance(getattr(event, 'payload', None), Exception):
            raise event.payload
        log.debug('unsupported event ignored: %s' % type(event))

    async def handler_event(self, sources, target, event):
        event.payload.set()

    async def handler_failed(self, sources, target, event):
        self.ndb.schema.mark(target, 1)

    def main(self):
        asyncio.run(self.run())

    def create_task(self, coro, state='running', obj=None):
        task = Task(coro, state, obj)
        self.task_map[task.task_id] = task
        self.reload_event.set()
        return task

    def restart_task(self, task):
        task.restart()
        self.task_map[task.task_id] = task
        self.reload_event.set()
        return task

    async def stop(self):
        await self.stop_event.wait()

    async def reload(self):
        await self.reload_event.wait()

    async def task_watch(self):
        while True:
            tasks = list([x.task for x in self.task_map.values()])
            done, pending = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED
            )
            self.log.debug(f'task done {done}')
            if self.stop_event.is_set():
                return
            for t in done:
                task = self.task_map.pop(id(t))
                task.commit()
                if task.target_state == 'running':
                    self.restart_task(task)
            self.reload_event.clear()

    async def receiver(self):
        while True:
            event = await self.event_queue.get()
            reschedule = []
            handlers = self.event_map.get(
                event.__class__, [self.handler_default]
            )

            for handler in tuple(handlers):
                try:
                    target = event['header']['target']
                    # self.log.debug(f'await {handler} for {event}')
                    await handler(self.ndb.sources.asyncore, target, event)
                except RescheduleException:
                    if 'rcounter' not in event['header']:
                        event['header']['rcounter'] = 0
                    if event['header']['rcounter'] < 3:
                        event['header']['rcounter'] += 1
                        self.log.debug('reschedule %s' % (event,))
                        reschedule.append(event)
                    else:
                        self.log.error('drop %s' % (event,))
                except InvalidateHandlerException:
                    try:
                        handlers.remove(handler)
                    except Exception:
                        self.log.error(
                            'could not invalidate '
                            'event handler:\n%s' % traceback.format_exc()
                        )
                except ShutdownException:
                    return
                except DBMExitException:
                    return
                except Exception:
                    self.log.error(
                        'could not load event:\n%s\n%s'
                        % (event, traceback.format_exc())
                    )
            if time.time() - self.gctime > config.gc_timeout:
                self.gctime = time.time()

    async def run(self):
        self.thread = id(threading.current_thread())

        # init the events map
        event_map = {
            cmsg_event: [self.handler_event],
            cmsg_failed: [self.handler_failed],
        }
        self.event_map = event_map

        try:
            self.ndb.schema = schema.DBSchema(
                self.ndb.config, self.event_map, self.log.channel('schema')
            )
            self.ndb.config = NDBConfig(self)

        except Exception as e:
            self.ndb._dbm_error = e
            self.ready.set()
            return

        for event, handlers in self.ndb.schema.event_map.items():
            for handler in handlers:
                self.register_handler(event, handler)

        # create an event loop
        self.event_loop = asyncio.get_event_loop()
        self.ndb.event_loop = self.event_loop
        self.create_task(self.receiver)
        self.create_task(self.stop)
        self.create_task(self.reload)
        self.ndb._dbm_ready.set()
        await self.task_watch()
        self.ndb.schema.close()
        self.ndb._dbm_shutdown.set()
        self.ready.set()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
from pyroute2.iwutil import IW

iw = IW()
for q in iw.get_interfaces_dump():
    phyname = 'phy%i' % int(q.get_attr('NL80211_ATTR_WIPHY'))
    print(
        '%s\t%s\t%s\t%s'
        % (
            q.get_attr('NL80211_ATTR_IFINDEX'),
            phyname,
            q.get_attr('NL80211_ATTR_IFNAME'),
            q.get_attr('NL80211_ATTR_MAC'),
        )
    )
iw.close()



from pyroute2.requests.main import RequestProcessor


class Request(dict):
    pass


class Result(dict):
    pass


def run_test(config, spec, result):
    processor = RequestProcessor(context=spec, prime=spec)
    for fspec in config['filters']:
        processor.add_filter(fspec['class'](*fspec['argv']))
    processor.finalize()
    assert Result(processor) == result



from socket import AF_INET, AF_INET6

from pyroute2.netlink import NLM_F_ACK, NLM_F_DUMP, NLM_F_REQUEST, genlmsg, nla
from pyroute2.netlink.generic import (
    AsyncGenericNetlinkSocket,
    GenericNetlinkSocket,
)

MPTCP_GENL_NAME = 'mptcp_pm'

MPTCP_PM_CMD_UNSPEC = 0
MPTCP_PM_CMD_ADD_ADDR = 1
MPTCP_PM_CMD_DEL_ADDR = 2
MPTCP_PM_CMD_GET_ADDR = 3
MPTCP_PM_CMD_FLUSH_ADDRS = 4
MPTCP_PM_CMD_SET_LIMITS = 5
MPTCP_PM_CMD_GET_LIMITS = 6
MPTCP_PM_CMD_SET_FLAGS = 7


class mptcp_msg(genlmsg):
    prefix = 'MPTCP_PM_ATTR_'
    nla_map = (
        ('MPTCP_PM_ATTR_UNSPEC', 'none'),
        ('MPTCP_PM_ATTR_ADDR', 'pm_addr'),
        ('MPTCP_PM_ATTR_RCV_ADD_ADDRS', 'uint32'),
        ('MPTCP_PM_ATTR_SUBFLOWS', 'uint32'),
    )

    class pm_addr(nla):
        prefix = 'MPTCP_PM_ADDR_ATTR_'
        nla_map = (
            ('MPTCP_PM_ADDR_ATTR_UNSPEC', 'none'),
            ('MPTCP_PM_ADDR_ATTR_FAMILY', 'uint16'),
            ('MPTCP_PM_ADDR_ATTR_ID', 'uint8'),
            ('MPTCP_PM_ADDR_ATTR_ADDR4', 'ipaddr'),
            ('MPTCP_PM_ADDR_ATTR_ADDR6', 'ipaddr'),
            ('MPTCP_PM_ADDR_ATTR_PORT', 'uint16'),
            ('MPTCP_PM_ADDR_ATTR_FLAGS', 'uint32'),
            ('MPTCP_PM_ADDR_ATTR_IF_IDX', 'hex'),
        )


class AsyncMPTCP(AsyncGenericNetlinkSocket):

    async def setup_endpoint(self):
        if getattr(self.local, 'transport', None) is not None:
            return
        await super().setup_endpoint()
        await self.bind(MPTCP_GENL_NAME, mptcp_msg)

    async def endpoint(self, cmd, **kwarg):
        '''
        Usage::

            mptcp.endpoint('show')
            mptcp.endpoint('add', addr='172.17.20.2')
            mptcp.endpoint('del', id=4)
            mptcp.endpoint('flush')

        Argument `addr` is equal to `addr4` and implies `family=AF_INET`,
        while `addr6` implies `family=AF_INET6`
        '''
        flags_dump = NLM_F_REQUEST | NLM_F_DUMP
        flags_base = NLM_F_REQUEST | NLM_F_ACK
        commands = {
            'show': (MPTCP_PM_CMD_GET_ADDR, flags_dump),
            'add': (MPTCP_PM_CMD_ADD_ADDR, flags_base),
            'del': (MPTCP_PM_CMD_DEL_ADDR, flags_base),
            'flush': (MPTCP_PM_CMD_FLUSH_ADDRS, flags_base),
        }

        (command, flags) = commands.get(cmd, cmd)
        msg = mptcp_msg()
        msg['cmd'] = command
        msg['version'] = 1

        if cmd in ('add', 'del'):
            addr_info = {'attrs': []}
            if 'addr' in kwarg:
                kwarg['addr4'] = kwarg.pop('addr')
            if 'addr4' in kwarg:
                kwarg['family'] = AF_INET
            elif 'addr6' in kwarg:
                kwarg['family'] = AF_INET6
            for key, value in kwarg.items():
                addr_info['attrs'].append(
                    (mptcp_msg.pm_addr.name2nla(key), value)
                )
            msg['attrs'] = [('MPTCP_PM_ATTR_ADDR', addr_info, 0x8000)]

        return await self.nlm_request(msg, msg_type=self.prid, msg_flags=flags)

    async def limits(self, cmd, **kwarg):
        '''
        Usage::

            mptcp.limits('show')
            mptcp.limits('set', subflows=10)
        '''
        flags_dump = NLM_F_REQUEST
        flags_base = NLM_F_REQUEST | NLM_F_ACK
        commands = {
            'show': (MPTCP_PM_CMD_GET_LIMITS, flags_dump),
            'set': (MPTCP_PM_CMD_SET_LIMITS, flags_base),
        }

        (command, flags) = commands.get(cmd, cmd)
        msg = mptcp_msg()
        msg['cmd'] = command
        msg['version'] = 1

        if cmd == 'set':
            if not set(kwarg) < set(
                ('subflows', 'rcv_add_addrs', 'add_addr_accepted')
            ):
                raise TypeError('invalid parameter')
            for key, value in kwarg.items():
                if key == 'add_addr_accepted':
                    key = 'rcv_add_addrs'
                msg['attrs'].append((mptcp_msg.name2nla(key), value))

        return await self.nlm_request(msg, msg_type=self.prid, msg_flags=flags)


class MPTCP(GenericNetlinkSocket):

    async_class = AsyncMPTCP

    def endpoint(self, cmd, **kwarg):
        return self._run_sync_cleanup(self.asyncore.endpoint, cmd, **kwarg)

    def limits(self, cmd, **kwarg):
        return self._run_sync_cleanup(self.asyncore.limits, cmd, **kwarg)



import asyncio
import logging
from socket import AF_PACKET, SOCK_RAW, SOL_SOCKET, errno, error, htons, socket
from typing import Optional

from pyroute2.ext import bpf
from pyroute2.iproute.linux import AsyncIPRoute
from pyroute2.netlink.rtnl import RTMGRP_LINK

LOG = logging.getLogger(__name__)
ETH_P_IP = 0x0800
ETH_P_ALL = 3
SO_ATTACH_FILTER = 26
SO_DETACH_FILTER = 27


total_filter = [[0x06, 0, 0, 0]]


class AsyncRawSocket(socket):
    '''
    This raw socket binds to an interface and optionally installs a BPF
    filter.
    When created, the socket's buffer is cleared to remove packets that
    arrived before bind() or the BPF filter is installed.  Doing so
    requires calling recvfrom() which may raise an exception if the
    interface is down.
    In order to allow creating the socket when the interface is
    down, the ENETDOWN exception is caught and discarded.
    '''

    fprog = None

    def __init__(self, ifname: str, bpf: Optional[list[list[int]]] = None):
        self.ifname = ifname
        self.bpf = bpf
        # start watching for mac addr changes
        self._l2addr_watcher: Optional[asyncio.Task] = None

    async def __aexit__(self, *_):
        self._l2addr_watcher.cancel()
        self.close()

    async def __aenter__(self):
        # lookup the interface details
        async with AsyncIPRoute() as ip:
            async for link in await ip.get_links():
                if link.get_attr('IFLA_IFNAME') == self.ifname:
                    break
            else:
                raise IOError(2, 'Link not found')
        self.l2addr: str = link.get_attr('IFLA_ADDRESS')
        self.ifindex: int = link['index']
        # bring up the socket
        socket.__init__(self, AF_PACKET, SOCK_RAW, htons(ETH_P_ALL))
        socket.setblocking(self, False)
        socket.bind(self, (self.ifname, ETH_P_ALL))
        if self.bpf:
            self.clear_buffer()
            fstring, self.fprog = bpf.compile(self.bpf)
            socket.setsockopt(self, SOL_SOCKET, SO_ATTACH_FILTER, fstring)
        else:
            # FIXME: should be async
            self.clear_buffer(remove_total_filter=True)
        # change self.l2addr if it changes
        self._l2addr_watcher = asyncio.create_task(
            self._watch_l2addr_changes(),
            name=f'Watch {self.ifname} for l2addr changes',
        )
        return self

    async def _watch_l2addr_changes(self):
        '''Updates self.l2addr when the interfaces's mac changes.

        During the lifetime of the socket, the interface's mac can change, and
        since it's read it at startup & used to build packets, they will then
        have the wrong mac.
        '''
        async with AsyncIPRoute() as ipr:
            await ipr.bind(RTMGRP_LINK)
            while True:
                async for msg in ipr.get():
                    if msg.get('IFLA_IFNAME') != self.ifname:
                        continue
                    new_l2addr = msg.get_attr('IFLA_ADDRESS')
                    if new_l2addr and new_l2addr != self.l2addr:
                        LOG.info(
                            'l2addr for %s changed from %s to %s',
                            self.ifname,
                            self.l2addr,
                            new_l2addr,
                        )
                        self.l2addr = new_l2addr

    def clear_buffer(self, remove_total_filter: bool = False):
        # there is a window of time after the socket has been created and
        # before bind/attaching a filter where packets can be queued onto the
        # socket buffer
        # see comments in function set_kernel_filter() in libpcap's
        # pcap-linux.c. libpcap sets a total filter which does not match any
        # packet.  It then clears what is already in the socket
        # before setting the desired filter
        total_fstring, prog = bpf.compile(total_filter)
        socket.setsockopt(self, SOL_SOCKET, SO_ATTACH_FILTER, total_fstring)
        while True:
            try:
                self.recvfrom(0)
            except error as e:
                if e.args[0] == errno.ENETDOWN:
                    # we only get this exception once per down event
                    # there may be more packets left to clean
                    pass
                elif e.args[0] in [errno.EAGAIN, errno.EWOULDBLOCK]:
                    break
                else:
                    raise
        if remove_total_filter:
            # total_fstring ignored
            socket.setsockopt(
                self, SOL_SOCKET, SO_DETACH_FILTER, total_fstring
            )

    @staticmethod
    def csum(data: bytes) -> int:
        '''Compute the "Internet checksum" for the given bytes.'''
        if len(data) % 2:
            data += b'\x00'
        csum: int = 0
        # pretty much the fastest way to compute this in Python
        for i in range(len(data) // 2):
            offset = i * 2
            csum += (data[offset] << 8) + data[offset + 1]
        csum = (csum >> 16) + (csum & 0xFFFF)
        csum += csum >> 16
        return ~csum & 0xFFFF



'''
hfsc
++++

Simple HFSC example::

    eth0 = ip.get_links(ifname="eth0")[0]
    ip.tc("add", "hfsc", eth0,
          handle="1:",
          default="1:1")
    ip.tc("add-class", "hfsc", eth0,
          handle="1:1",
          parent="1:0"
          rsc={"m2": "5mbit"})

HFSC curve nla types:

* `rsc`: real-time curve
* `fsc`: link-share curve
* `usc`: upper-limit curve
'''

from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import RTM_DELQDISC, RTM_NEWQDISC, TC_H_ROOT
from pyroute2.netlink.rtnl.tcmsg.common import get_rate, get_time
from pyroute2.netlink.rtnl.tcmsg.common import stats2 as c_stats2

parent = TC_H_ROOT


def get_parameters(kwarg):
    defcls = kwarg.get('default', kwarg.get('defcls', 0x10))
    defcls &= 0xFFFF
    return {'defcls': defcls}


def get_class_parameters(kwarg):
    ret = {'attrs': []}
    for key in ('rsc', 'fsc', 'usc'):
        if key in kwarg:
            ret['attrs'].append(
                [
                    'TCA_HFSC_%s' % key.upper(),
                    {
                        'm1': get_rate(kwarg[key].get('m1', 0)),
                        'd': get_time(kwarg[key].get('d', 0)),
                        'm2': get_rate(kwarg[key].get('m2', 0)),
                    },
                ]
            )
    return ret


class options_hfsc(nla):
    fields = (('defcls', 'H'),)  # default class


class options_hfsc_class(nla):
    nla_map = (
        ('TCA_HFSC_UNSPEC', 'none'),
        ('TCA_HFSC_RSC', 'hfsc_curve'),  # real-time curve
        ('TCA_HFSC_FSC', 'hfsc_curve'),  # link-share curve
        ('TCA_HFSC_USC', 'hfsc_curve'),
    )  # upper-limit curve

    class hfsc_curve(nla):
        fields = (
            ('m1', 'I'),  # slope of the first segment in bps
            ('d', 'I'),  # x-projection of the first segment in us
            ('m2', 'I'),
        )  # slope of the second segment in bps


def options(msg, *argv, **kwarg):
    if msg['header']['type'] in (RTM_NEWQDISC, RTM_DELQDISC):
        return options_hfsc
    else:
        return options_hfsc_class


class stats2(c_stats2):
    class stats_app(nla):
        fields = (
            ('work', 'Q'),  # total work done
            ('rtwork', 'Q'),  # total work done by real-time criteria
            ('period', 'I'),  # current period
            ('level', 'I'),
        )  # class level in hierarchy



#!/usr/bin/env python3

import traceback
from pprint import pprint

from pyroute2.netlink.generic import GenericNetlinkSocket

if __name__ == '__main__':
    try:
        # create protocol instance
        genl = GenericNetlinkSocket(ext_ack=True)

        # extract policy
        msg = genl.policy('nlctrl')

        # dump policy information
        pprint(msg)
    except:
        # if there was an error, log it to the console
        traceback.print_exc()
    finally:
        # finally -- release the instance
        genl.close()



from pyroute2.netlink import nla


class vti6(nla):
    prefix = 'IFLA_'
    nla_map = (
        ('IFLA_VTI_UNSPEC', 'none'),
        ('IFLA_VTI_LINK', 'uint32'),
        ('IFLA_VTI_IKEY', 'be32'),
        ('IFLA_VTI_OKEY', 'be32'),
        ('IFLA_VTI_LOCAL', 'ip6addr'),
        ('IFLA_VTI_REMOTE', 'ip6addr'),
        ('IFLA_VTI_FWMARK', 'uint32'),
    )



import errno

from pyroute2 import IW, IPRoute
from pyroute2.netlink.exceptions import NetlinkError

# interface name to check
ifname = 'lo'

ip = IPRoute()
iw = IW()
index = ip.link_lookup(ifname=ifname)[0]
try:
    iw.get_interface_by_ifindex(index)
    print("wireless interface")
except NetlinkError as e:
    if e.code == errno.ENODEV:  # 19 'No such device'
        print("not a wireless interface")
finally:
    iw.close()
    ip.close()



from collections.abc import Generator

import pytest

from pyroute2 import NDB


@pytest.fixture
def ndb(nsname: str) -> Generator[NDB]:
    '''NDB instance.

    * **Name**: ndb
    * **Scope**: function
    * **Depends**: nsname

    Yield `NDB` instance running in the test network namespace.
    '''
    with NDB(sources=[{'target': 'localhost', 'netns': nsname}]) as ndb:
        yield ndb



import pytest

from pyroute2.netlink import nlmsg

prime = {
    'attrs': (
        ('A', 2),
        ('A', 3),
        ('A', 4),
        ('B', {'attrs': (('C', 5), ('D', {'attrs': (('E', 6), ('F', 7))}))}),
    )
}


@pytest.fixture
def msg():
    msg = nlmsg()
    msg.setvalue(prime)
    yield msg


def test_get_attr(msg):
    assert msg.get_attr('A') == 2
    assert msg.get_attr('C') is None


def test_get_attrs(msg):
    assert msg.get_attrs('A') == [2, 3, 4]
    assert msg.get_attrs('C') == []


def test_get_nested(msg):
    assert msg.get_nested('B', 'D', 'E') == 6
    assert msg.get_nested('B', 'D', 'F') == 7
    assert msg.get_nested('B', 'D', 'G') is None
    assert msg.get_nested('C', 'D', 'E') is None



import asyncio
import sys
from typing import Literal, Optional

from pyroute2.iproute.linux import AsyncIPRoute
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.rtnl import RTMGRP_LINK

IfaceState = Literal['up', 'down']


class InterfaceNotFound(LookupError):
    '''Raised when an interface is not found.'''


class InterfaceStateWatcher:
    '''Async context manager, fires events when an interface changes state.

    Used by the dhcp client to restart itself in such cases.
    '''

    def __init__(self, interface: str) -> None:
        self.interface = interface
        self.up = asyncio.Event()
        self.down = asyncio.Event()
        self._state: Optional[IfaceState] = None
        self._watcher: Optional[asyncio.Task] = None
        self._ipr: Optional[AsyncIPRoute] = None

    @property
    def ipr(self) -> AsyncIPRoute:
        '''The async iproute context. Only useable in the context manager.'''
        assert self._ipr, 'need to use as a context manager'
        return self._ipr

    @property
    def state(self) -> IfaceState:
        '''The current state of the watched interface.'''
        assert self._state, "not yet started"
        return self._state

    @state.setter
    def state(self, value: IfaceState) -> None:
        '''Set the state & trigger the relevant event.'''
        self._state = value
        if value == 'up':
            self.up.set()
            self.down.clear()
        elif value == 'down':
            self.down.set()
            self.up.clear()

    async def _fetch_current_state(self) -> IfaceState:
        '''Get the initial state before we're notified of changes.'''
        lookup_results = await self.ipr.link_lookup(ifname=self.interface)
        if not lookup_results:
            raise InterfaceNotFound(self.interface)
        get_results = await self.ipr.link("get", index=lookup_results[0])
        return get_results[0].get('state')

    async def _watch_changes(self) -> None:
        '''Updates `state` in real time forever.'''
        while True:
            # TODO: svinota says we should call .link('dump') here
            # to empty the buffer before starting
            async for msg in self.ipr.get():
                if msg.get('IFLA_IFNAME') == self.interface:
                    self.state = msg.get('state')

    async def __aenter__(self) -> 'InterfaceStateWatcher':
        self._ipr = await AsyncIPRoute().__aenter__()
        await self.ipr.bind(RTMGRP_LINK)
        self.state = await self._fetch_current_state()
        self._watcher = asyncio.create_task(self._watch_changes())
        return self

    async def __aexit__(self, *_) -> None:
        self.ipr.close()
        if self._watcher:
            try:
                await self._watcher
            except NetlinkError as exc:
                if exc.code == 104:
                    # we called close() so that is expected
                    pass
                else:
                    raise
            self._watcher = None
        self._ipr = None


async def main(iface: str, state: str) -> None:
    '''Very basic entrypoint for commandline testing.'''
    assert state in ('up', 'down')
    async with InterfaceStateWatcher(iface) as watcher:
        print('Will exit when', iface, 'is', state)
        await getattr(watcher, state).wait()


if __name__ == '__main__':
    try:
        asyncio.run(main(sys.argv[1], sys.argv[2]))
    except IndexError:
        print('Usage:', sys.argv[0], '<interface> <status>')



import builtins
import json
import struct

from pyroute2.netlink import nlmsg
from pyroute2.netlink.nlsocket import Marshal

Tversion = 100
Rversion = 101
Tauth = 102
Rauth = 103
Tattach = 104
Rattach = 105
Terror = 106  # illegal
Rerror = 107
Tflush = 108
Rflush = 109
Twalk = 110
Rwalk = 111
Topen = 112
Ropen = 113
Tcreate = 114
Rcreate = 115
Tread = 116
Rread = 117
Twrite = 118
Rwrite = 119
Tclunk = 120
Rclunk = 121
Tremove = 122
Rremove = 123
Tstat = 124
Rstat = 125
Twstat = 126
Rwstat = 127
Topenfd = 98
Ropenfd = 99

# 9P2000.pr2 extensions
Tcall = 80
Rcall = 81


class Plan9Exit(Exception):
    pass


def array(kind, header='H'):
    class CustomArray:

        @staticmethod
        def decode_count(data, offset):
            (count,) = struct.unpack_from(header, data, offset)
            return count, offset + struct.calcsize(header)

        @staticmethod
        def decode_from(data, offset):
            count, offset = CustomArray.decode_count(data, offset)
            ret = []
            for _ in range(count):
                value, offset = kind.decode_from(data, offset)
                ret.append(value)
            return ret, offset

        @staticmethod
        def encode_into(data, offset, value):
            if not isinstance(value, (tuple, list)):
                value = []
            data.extend([0] * struct.calcsize(header))
            struct.pack_into(header, data, offset, len(value))
            offset += struct.calcsize(header)
            for item in value:
                offset = kind.encode_into(data, offset, item)
            return offset

    return CustomArray


class Qid(dict):
    length = 13

    def __init__(self, qtype, vers, path):
        self['type'] = qtype
        self['vers'] = vers
        self['path'] = path

    @staticmethod
    def decode_from(data, offset):
        return (
            dict(
                zip(
                    ('type', 'vers', 'path'),
                    struct.unpack_from('=BIQ', data, offset),
                )
            ),
            offset + struct.calcsize('=BIQ'),
        )

    @staticmethod
    def encode_into(data, offset, value):
        data.extend([0] * Qid.length)
        struct.pack_into(
            '=BIQ', data, offset, value['type'], value['vers'], value['path']
        )
        return offset + Qid.length


class Stat(dict):
    header_fmt = 'H'

    def __init__(self):
        self['size'] = 58
        self['type'] = 0
        self['dev'] = 0
        self['qid.type'] = 0
        self['qid.vers'] = 0
        self['qid.path'] = 0
        self['mode'] = 0
        self['atime'] = 0
        self['mtime'] = 0
        self['length'] = 0
        self['name'] = ''
        self['uid'] = ''
        self['gid'] = ''
        self['muid'] = ''

    @staticmethod
    def decode_from(data, offset=0):
        ret = dict(
            zip(
                (
                    'size',
                    'type',
                    'dev',
                    'qid.type',
                    'qid.vers',
                    'qid.path',
                    'mode',
                    'atime',
                    'mtime',
                    'length',
                ),
                struct.unpack_from('=HHIBIQIIIQ', data, offset),
            )
        )
        offset += 41
        for key in ('name', 'uid', 'gid', 'muid'):
            ret[key], offset = String.decode_from(data, offset)
        return ret, offset

    @staticmethod
    def encode_into(data, offset, value):
        data.extend([0] * 41)
        # size of all the data except uint16 `size` header
        # 41 + 2 + name + 2 + uid + 2 + gid + 2 + muid
        value['size'] = (
            47
            + len(value['name'])
            + len(value['uid'])
            + len(value['gid'])
            + len(value['muid'])
        )
        struct.pack_into(
            '=HHIBIQIIIQ',
            data,
            offset,
            value['size'],
            value['type'],
            value['dev'],
            value['qid.type'],
            value['qid.vers'],
            value['qid.path'],
            value['mode'],
            value['atime'],
            value['mtime'],
            value['length'],
        )
        offset += 41
        for key in ('name', 'uid', 'gid', 'muid'):
            offset = String.encode_into(data, offset, value[key])
        return offset


class WStat(Stat):
    @staticmethod
    def decode_from(data, offset=0):
        # just ignore plength for now
        return Stat.decode_from(data, offset + 2)

    @staticmethod
    def encode_into(data, offset, value):
        data.extend([0] * 2)
        new_offset = Stat.encode_into(data, offset + 2, value)
        # size of all the data except uint16 `plength` header
        struct.pack_into('H', data, offset, new_offset - offset - 2)
        return new_offset


class CData:
    header_fmt = 'I'

    @staticmethod
    def decode_from(data, offset=0):
        (length,) = struct.unpack_from(CData.header_fmt, data, offset)
        offset += struct.calcsize(CData.header_fmt)
        return bytearray(data[offset : offset + length]), offset + length

    @staticmethod
    def encode_into(data, offset, value):
        length = len(value)
        if isinstance(value, str):
            value = value.encode('utf-8')
        data.extend([0] * (length + struct.calcsize(CData.header_fmt)))
        struct.pack_into(
            f'{CData.header_fmt}{length}s', data, offset, length, value
        )
        return offset + length + struct.calcsize(CData.header_fmt)


class String:
    header_fmt = 'H'

    @staticmethod
    def decode_from(data, offset=0):
        (length,) = struct.unpack_from(String.header_fmt, data, offset)
        offset += struct.calcsize(String.header_fmt)
        (value,) = struct.unpack_from(f'{length}s', data, offset)
        value = value.decode('utf-8')
        return value, offset + length

    @staticmethod
    def encode_into(data, offset, value):
        length = len(value)
        data.extend([0] * (length + struct.calcsize(String.header_fmt)))
        struct.pack_into(
            f'{String.header_fmt}{length}s',
            data,
            offset,
            length,
            value.encode('utf-8'),
        )
        return offset + length + struct.calcsize(String.header_fmt)


class msg_base(nlmsg):
    align = 0
    header = (('length', '=I'), ('type', 'B'), ('tag', 'H'))


class msg_terror(msg_base):
    defaults = {'header': {'type': Terror}}


class msg_rerror(msg_base):
    defaults = {'header': {'type': Rerror}}
    fields = (('ename', String),)


class msg_tversion(msg_base):
    defaults = {'header': {'type': Tversion}}
    fields = (('msize', 'I'), ('version', String))


class msg_rversion(msg_base):
    defaults = {'header': {'type': Rversion}}
    fields = (('msize', 'I'), ('version', String))


class msg_tauth(msg_base):
    defaults = {'header': {'type': Tauth}}
    fields = (('afid', 'I'), ('uname', String), ('aname', String))


class msg_rauth(msg_base):
    defaults = {'header': {'type': Rauth}}
    fields = (('aqid', '13B'),)


class msg_tattach(msg_base):
    defaults = {'header': {'type': Tattach}}
    fields = (
        ('fid', 'I'),
        ('afid', 'I'),
        ('uname', String),
        ('aname', String),
    )


class msg_rattach(msg_base):
    defaults = {'header': {'type': Rattach}}
    fields = (('qid', Qid),)


class msg_twalk(msg_base):
    defaults = {'header': {'type': Twalk}}
    fields = (('fid', 'I'), ('newfid', 'I'), ('wname', array(String)))


class msg_rwalk(msg_base):
    defaults = {'header': {'type': Rwalk}}
    fields = (('wqid', array(Qid)),)


class msg_tstat(msg_base):
    defaults = {'header': {'type': Tstat}}
    fields = (('fid', 'I'),)


class msg_rstat(msg_base):
    defaults = {'header': {'type': Rstat}}
    fields = (('stat', WStat),)


class msg_twstat(msg_base):
    defaults = {'header': {'type': Twstat}}
    fields = (('fid', 'I'), ('stat', WStat))


class msg_rwstat(msg_base):
    defaults = {'header': {'type': Rwstat}}


class msg_tclunk(msg_base):
    defaults = {'header': {'type': Tclunk}}
    fields = (('fid', 'I'),)


class msg_rclunk(msg_base):
    defaults = {'header': {'type': Rclunk}}
    pass


class msg_topen(msg_base):
    defaults = {'header': {'type': Topen}}
    fields = (('fid', 'I'), ('mode', 'B'))


class msg_ropen(msg_base):
    defaults = {'header': {'type': Ropen}}
    fields = (('qid', Qid), ('iounit', 'I'))


class msg_tread(msg_base):
    defaults = {'header': {'type': Tread}}
    fields = (('fid', 'I'), ('offset', 'Q'), ('count', 'I'))


class msg_rread(msg_base):
    defaults = {'header': {'type': Rread}, 'data': b''}
    fields = (('data', CData),)


class msg_twrite(msg_base):
    defaults = {'header': {'type': Twrite}}
    fields = (('fid', 'I'), ('offset', 'Q'), ('data', CData))


class msg_rwrite(msg_base):
    defaults = {'header': {'type': Rwrite}}
    fields = (('count', 'I'),)


class msg_tcall(msg_base):
    defaults = {'header': {'type': Tcall}}
    fields = (('fid', 'I'), ('text', String), ('data', CData))


class msg_rcall(msg_base):
    defaults = {'header': {'type': Rcall}}
    fields = (('err', 'H'), ('text', String), ('data', CData))


class Marshal9P(Marshal):
    default_message_class = msg_rerror

    msg_map = {
        Tversion: msg_tversion,
        Rversion: msg_rversion,
        Tauth: msg_tauth,
        Rauth: msg_rauth,
        Tattach: msg_tattach,
        Rattach: msg_rattach,
        Rerror: msg_rerror,
        Twalk: msg_twalk,
        Rwalk: msg_rwalk,
        Topen: msg_topen,
        Ropen: msg_ropen,
        Tread: msg_tread,
        Rread: msg_rread,
        Tclunk: msg_tclunk,
        Rclunk: msg_rclunk,
        Tstat: msg_tstat,
        Rstat: msg_rstat,
        Twstat: msg_twstat,
        Rwstat: msg_rwstat,
        Twrite: msg_twrite,
        Rwrite: msg_rwrite,
        Tcall: msg_tcall,
        Rcall: msg_rcall,
        Tcreate: msg_base,
        Rcreate: msg_base,
        Tremove: msg_base,
        Rremove: msg_base,
    }

    def parse(self, data, seq=None, callback=None, skip_alien_seq=False):
        offset = 0
        while offset <= len(data) - 5:
            (length, key, tag) = struct.unpack_from('=IBH', data, offset)
            if skip_alien_seq and tag != seq:
                continue
            if not 0 < length <= len(data):
                break
            parser = self.get_parser(key, 0, tag)
            msg = parser(data, offset, length)
            if key == Rerror:
                spec = json.loads(msg['ename'])
                if spec['class'] in dir(builtins):
                    cls = getattr(builtins, spec['class'])
                elif spec['class'] == 'Plan9Exit':
                    cls = Plan9Exit
                else:
                    cls = Exception
                if not spec.get('argv'):
                    spec['argv'] = [spec['str']]
                raise cls(*spec['argv'])
            offset += length
            if msg is None:
                continue
            yield msg



import asyncio
import logging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from importlib import import_module
from typing import Any, Optional

from pyroute2.dhcp.client import AsyncDHCPClient, ClientConfig
from pyroute2.dhcp.fsm import State
from pyroute2.dhcp.hooks import Hook
from pyroute2.dhcp.iface_status import InterfaceNotFound, InterfaceStateWatcher
from pyroute2.dhcp.leases import Lease

LOG = logging.getLogger(__name__)


def import_dotted_name(name: str) -> Any:
    '''Import anything by name. Return None if the import wasn't successful.'''
    try:
        module_name, obj_name = name.rsplit('.', 1)
        module = import_module(module_name)
        return getattr(module, obj_name)
    except (ValueError, ImportError, AttributeError):
        return None


def get_psr() -> ArgumentParser:
    psr = ArgumentParser(
        description='A DHCP client based on pyroute2. '
        'Tries to obtain & keep a lease on an interface, running '
        'configurable hooks to assign the obtained IP address and gw to it.',
        epilog='Send a SIGUSR1 to renew the current lease, SIGUSR2 to rebind '
        'it and SIGHUP to reset & get a new lease.',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    psr.add_argument(
        'interface', help='The interface to request an address for.'
    )
    psr.add_argument(
        '--lease-type',
        help='Class to use for leases. '
        'Must be a subclass of `pyroute2.dhcp.leases.Lease`.',
        type=str,
        default='pyroute2.dhcp.leases.JSONFileLease',
        metavar='dotted.name',
    )
    psr.add_argument(
        '--hook',
        help='Hooks to load. '
        'These are used to run async python code when, '
        'for example, renewing or expiring a lease. '
        'Defaults to adding & removing ip & gateway.',
        action='append',
        type=str,
        metavar='dotted.name',
    )
    psr.add_argument(
        '--disable-hooks',
        help='Disable all hooks.',
        default=False,
        action='store_true',
    )
    psr.add_argument(
        '-x',
        '--exit-on-timeout',
        metavar='N',
        help='Wait for max N seconds for a lease, '
        'exit if none could be obtained.',
        type=float,
    )
    psr.add_argument(
        '--log-level',
        help='Logging level to use.',
        choices=('DEBUG', 'INFO', 'WARNING', 'ERROR'),
        default='INFO',
    )
    psr.add_argument(
        '-p',
        '--write-pidfile',
        default=False,
        action='store_true',
        help='Write a pid file in the working directory. ',
    )
    psr.add_argument(
        '-R',
        '--no-release',
        default=False,
        action='store_true',
        help='Do not send a DHCPRELEASE on exit.',
    )
    # TODO: add options for parameters, retransmission, timeouts...
    return psr


async def run_client(
    cfg: ClientConfig, exit_timeout: Optional[float] = None
) -> None:
    '''Run the client until interrupted, or a timeout occurs.

    The optional `exit_timeout` controls 2 things when provided:
    - How long to wait for the interface to be up
    - How long to wait for the client to be bound when starting up
    '''

    acli = AsyncDHCPClient(cfg)

    async with InterfaceStateWatcher(cfg.interface) as iface_watcher:
        while True:
            # Open the socket, read existing lease, etc
            if iface_watcher.state != 'up':
                LOG.info('Waiting for %s to go up...', cfg.interface)
            await asyncio.wait_for(
                iface_watcher.up.wait(), timeout=exit_timeout
            )
            async with acli:
                # Bootstrap the client by sending a DISCOVER or a REQUEST
                await acli.bootstrap()
                if exit_timeout:
                    # Wait a bit for a lease, and raise if we have none
                    await acli.wait_for_state(
                        State.BOUND, timeout=exit_timeout
                    )
                    break
                await iface_watcher.down.wait()
                LOG.warning('%s went down', cfg.interface)


async def main() -> None:
    psr = get_psr()
    args = psr.parse_args()
    logging.basicConfig(
        format='%(asctime)s %(levelname)s [%(name)s:%(funcName)s] %(message)s'
    )
    logging.getLogger('pyroute2.dhcp').setLevel(args.log_level)

    LOG.setLevel(args.log_level)

    # parse lease type
    lease_type = import_dotted_name(args.lease_type)
    if not (isinstance(lease_type, type) and issubclass(lease_type, Lease)):
        psr.error(f'{args.lease_type!r} must point to a Lease subclass.')

    # parse hooks
    hooks: list[Hook] = []
    if not args.disable_hooks:
        if not args.hook:
            args.hook = [
                'pyroute2.dhcp.hooks.configure_ip',
                'pyroute2.dhcp.hooks.add_default_gw',
                'pyroute2.dhcp.hooks.remove_default_gw',
                'pyroute2.dhcp.hooks.remove_ip',
            ]
        LOG.debug('Configured hooks:')
        for dotted_hook_name in args.hook:
            hook = import_dotted_name(dotted_hook_name)
            if not isinstance(hook, Hook):
                psr.error(f'{dotted_hook_name!r} must point to a valid hook.')
            hooks.append(hook)
            LOG.debug("- %s", hook.name)

    # Create configuration
    cfg = ClientConfig(
        interface=args.interface,
        lease_type=lease_type,
        hooks=hooks,
        write_pidfile=args.write_pidfile,
        release=not args.no_release,
        handle_signals=True,
    )

    try:
        await run_client(cfg, exit_timeout=args.exit_on_timeout)
    except InterfaceNotFound as err:
        psr.error(f"Interface not found: {err}")


def run():
    # for the setup.cfg entrypoint
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':  # pragma: no cover
    run()



import time
from collections.abc import AsyncGenerator
from socket import socketpair

import pytest_asyncio

from pyroute2.plan9.client import Plan9ClientSocket
from pyroute2.plan9.server import Plan9ServerSocket


def test_time():
    return time.time_ns()


class AsyncPlan9Context:

    server = None
    client = None
    shutdown_response = None
    sample_data = b'Pi6raTaXuzohdu7n'

    def __init__(self):
        self.server_sock, self.client_sock = socketpair()
        self.server = Plan9ServerSocket(use_socket=self.server_sock)
        self.client = Plan9ClientSocket(use_socket=self.client_sock)
        self._task = None
        with self.server.filesystem.create('test_file') as i:
            i.data.write(self.sample_data)
        with self.server.filesystem.create('test_time') as i:
            i.metadata.call_on_read = True
            i.register_function(test_time, loader=lambda x: {})

    async def ensure_session(self):
        self._task = await self.server.async_run()
        await self.client.start_session()

    def close(self):
        self._task.cancel()
        self.client.close()
        self.server.close()


@pytest_asyncio.fixture
async def async_p9_context() -> AsyncGenerator[AsyncPlan9Context]:
    ctx = AsyncPlan9Context()
    await ctx.ensure_session()
    yield ctx
    ctx.close()



import struct
from socket import AF_INET, AF_INET6, AF_UNSPEC, inet_ntop, inet_pton

from pyroute2.common import AF_MPLS, hexdump, map_namespace
from pyroute2.netlink import nla, nla_string, nlmsg
from pyroute2.netlink.rtnl import rt_proto, rt_scope, rt_type

RTNH_F_DEAD = 1
RTNH_F_PERVASIVE = 2
RTNH_F_ONLINK = 4
RTNH_F_OFFLOAD = 8
RTNH_F_LINKDOWN = 16
(RTNH_F_NAMES, RTNH_F_VALUES) = map_namespace('RTNH_F_', globals())

RT_SCOPE_UNIVERSE = 0
RT_SCOPE_SITE = 200
RT_SCOPE_LINK = 253
RT_SCOPE_HOST = 254
RT_SCOPE_NOWHERE = 255
(RT_SCOPE_NAMES, RT_SCOPE_VALUES) = map_namespace('RT_SCOPE_', globals())

LWTUNNEL_ENCAP_NONE = 0
LWTUNNEL_ENCAP_MPLS = 1
LWTUNNEL_ENCAP_IP = 2
LWTUNNEL_ENCAP_ILA = 3
LWTUNNEL_ENCAP_IP6 = 4
LWTUNNEL_ENCAP_SEG6 = 5
LWTUNNEL_ENCAP_BPF = 6
LWTUNNEL_ENCAP_SEG6_LOCAL = 7

IP6_RT_PRIO_USER = 1024


class nlflags(object):
    def encode(self):
        if isinstance(self['flags'], str):
            self['flags'] = self['flags'].split(',')
        if isinstance(self['flags'], (set, tuple, list)):
            self['flags'] = self.names2flags(self['flags'])
        if isinstance(self.get('scope'), str):
            self['scope'] = self.name2scope(self['scope'])
        return super(nlflags, self).encode()

    @staticmethod
    def scope2name(scope):
        return RT_SCOPE_VALUES[scope].lower()[9:]

    @staticmethod
    def name2scope(scope):
        return RT_SCOPE_NAMES['RT_SCOPE_' + scope.upper()]

    @staticmethod
    def flags2names(flags):
        ret = []
        for flag in RTNH_F_VALUES:
            if (flag & flags) == flag:
                ret.append(RTNH_F_VALUES[flag].lower()[7:])
        return ret

    @staticmethod
    def names2flags(flags):
        ret = 0
        if isinstance(flags, str):
            flags = [flags]
        for flag in flags:
            ret |= RTNH_F_NAMES['RTNH_F_' + flag.upper()]
        return ret


class rtmsg_base(nlflags):
    '''
    Route message

    C structure::

        struct rtmsg {
            unsigned char rtm_family;   /* Address family of route */
            unsigned char rtm_dst_len;  /* Length of destination */
            unsigned char rtm_src_len;  /* Length of source */
            unsigned char rtm_tos;      /* TOS filter */

            unsigned char rtm_table;    /* Routing table ID */
            unsigned char rtm_protocol; /* Routing protocol; see below */
            unsigned char rtm_scope;    /* See below */
            unsigned char rtm_type;     /* See below */

            unsigned int  rtm_flags;
        };
    '''

    __slots__ = ()

    prefix = 'RTA_'
    sql_constraints = {
        'RTA_TABLE': 'NOT NULL DEFAULT 0',
        'RTA_DST': "NOT NULL DEFAULT ''",
        'RTA_OIF': 'NOT NULL DEFAULT 0',
        'RTA_PRIORITY': 'NOT NULL DEFAULT 0',
        'RTA_VIA': "NOT NULL DEFAULT ''",
        'RTA_NEWDST': "NOT NULL DEFAULT ''",
    }

    defaults = {
        'family': AF_INET,
        'scope': rt_scope['universe'],
        'type': rt_type['unspec'],
        'proto': rt_proto['unspec'],
    }

    field_policies = {('field', 'table'): lambda x: 252 if x > 255 else x}

    fields = (
        ('family', 'B'),
        ('dst_len', 'B'),
        ('src_len', 'B'),
        ('tos', 'B'),
        ('table', 'B'),
        ('proto', 'B'),
        ('scope', 'B'),
        ('type', 'B'),
        ('flags', 'I'),
    )

    nla_map = (
        ('RTA_UNSPEC', 'none'),
        ('RTA_DST', 'target'),
        ('RTA_SRC', 'target'),
        ('RTA_IIF', 'uint32'),
        ('RTA_OIF', 'uint32'),
        ('RTA_GATEWAY', 'target'),
        ('RTA_PRIORITY', 'uint32'),
        ('RTA_PREFSRC', 'target'),
        ('RTA_METRICS', 'metrics'),
        ('RTA_MULTIPATH', '*get_nh'),
        ('RTA_PROTOINFO', 'uint32'),
        ('RTA_FLOW', 'uint32'),
        ('RTA_CACHEINFO', 'cacheinfo'),
        ('RTA_SESSION', 'hex'),
        ('RTA_MP_ALGO', 'hex'),
        ('RTA_TABLE', 'uint32'),
        ('RTA_MARK', 'uint32'),
        ('RTA_MFC_STATS', 'rta_mfc_stats'),
        ('RTA_VIA', 'rtvia'),
        ('RTA_NEWDST', 'target'),
        ('RTA_PREF', 'uint8'),
        ('RTA_ENCAP_TYPE', 'uint16'),
        ('RTA_ENCAP', 'encap_info'),
        ('RTA_EXPIRES', 'hex'),
    )

    @staticmethod
    def encap_info(self, *argv, **kwarg):
        encap_type = None

        # Check, if RTA_ENCAP_TYPE is decoded already
        #
        for name, value in self['attrs']:
            if name == 'RTA_ENCAP_TYPE':
                encap_type = value
                break
        else:
            # No RTA_ENCAP_TYPE met, so iterate all the chain.
            # Ugly, but to do otherwise would be too complicated.
            #
            data = kwarg['data']
            offset = kwarg['offset']
            while offset < len(data):
                # Shift offset to the next NLA
                # NLA header:
                #
                # uint16 length
                # uint16 type
                #
                try:
                    offset += struct.unpack('H', data[offset : offset + 2])[0]
                    # 21 == RTA_ENCAP_TYPE
                    # FIXME: should not be hardcoded
                    if (
                        struct.unpack('H', data[offset + 2 : offset + 4])[0]
                        == 21
                    ):
                        encap_type = struct.unpack(
                            'H', data[offset + 4 : offset + 6]
                        )[0]
                        break
                except:
                    # in the case of any decoding error return self.hex
                    break

        # return specific classes
        #
        return self.encaps.get(encap_type, self.hex)

    class mpls_encap_info(nla):
        prefix = 'MPLS_IPTUNNEL_'
        __slots__ = ()

        nla_map = (
            ('MPLS_IPTUNNEL_UNSPEC', 'none'),
            ('MPLS_IPTUNNEL_DST', 'mpls_target'),
            ('MPLS_IPTUNNEL_TTL', 'uint8'),
        )

    class seg6_encap_info(nla):
        __slots__ = ()

        nla_map = (
            ('SEG6_IPTUNNEL_UNSPEC', 'none'),
            ('SEG6_IPTUNNEL_SRH', 'ipv6_sr_hdr'),
        )

        class ipv6_sr_hdr(nla):
            __slots__ = ()

            fields = (
                ('encapmode', 'I'),
                ('nexthdr', 'B'),
                ('hdrlen', 'B'),
                ('type', 'B'),
                ('segments_left', 'B'),
                ('first_segment', 'B'),
                ('flags', 'B'),
                ('reserved', 'H'),
                ('segs', 's'),
                # Potentially several type-length-value
                ('tlvs', 's'),
            )

            # Corresponding values for seg6 encap modes
            SEG6_IPTUN_MODE_INLINE = 0
            SEG6_IPTUN_MODE_ENCAP = 1

            # Mapping string to nla value
            encapmodes = {
                "inline": SEG6_IPTUN_MODE_INLINE,
                "encap": SEG6_IPTUN_MODE_ENCAP,
            }

            # Reverse mapping: mapping nla value to string
            r_encapmodes = {v: k for k, v in encapmodes.items()}

            # Nla value for seg6 type
            SEG6_TYPE = 4

            # Flag value for hmac
            SR6_FLAG1_HMAC = 1 << 3

            # Tlv value for hmac
            SR6_TLV_HMAC = 5

            # Utility function to get the family from the msg
            def get_family(self):
                pointer = self
                while pointer.parent is not None:
                    pointer = pointer.parent
                return pointer.get('family', AF_UNSPEC)

            def encode(self):
                # Retrieve the family
                family = self.get_family()
                # Seg6 can be applied only to IPv6 and IPv4
                if family == AF_INET6 or family == AF_INET:
                    # Get mode
                    mode = self['mode']
                    # Get segs
                    segs = self['segs']
                    # Get hmac
                    hmac = self.get('hmac', None)
                    # With "inline" mode there is not
                    # encap into an outer IPv6 header
                    if mode == "inline":
                        # Add :: to segs
                        segs.insert(0, "::")
                    # Add mode to value
                    self['encapmode'] = self.encapmodes.get(
                        mode, self.SEG6_IPTUN_MODE_ENCAP
                    )
                    # Calculate srlen
                    srhlen = 8 + 16 * len(segs)
                    # If we are using hmac we have a tlv as trailer data
                    if hmac:
                        # Since we can use sha1 or sha256
                        srhlen += 40
                    # Calculate and set hdrlen
                    self['hdrlen'] = (srhlen >> 3) - 1
                    # Add seg6 type
                    self['type'] = self.SEG6_TYPE
                    # Add segments left
                    self['segments_left'] = len(segs) - 1
                    # Add fitst segment
                    self['first_segment'] = len(segs) - 1
                    # If hmac is used we have to set the flags
                    if hmac:
                        # Add SR6_FLAG1_HMAC
                        self['flags'] |= self.SR6_FLAG1_HMAC
                    # Init segs
                    self['segs'] = b''
                    # Iterate over segments
                    for seg in segs:
                        # Convert to network byte order and add to value
                        self['segs'] += inet_pton(AF_INET6, seg)
                    # Initialize tlvs
                    self['tlvs'] = b''
                    # If hmac is used we have to properly init tlvs
                    if hmac:
                        # Put type
                        self['tlvs'] += struct.pack('B', self.SR6_TLV_HMAC)
                        # Put length -> 40-2
                        self['tlvs'] += struct.pack('B', 38)
                        # Put reserved
                        self['tlvs'] += struct.pack('H', 0)
                        # Put hmac key
                        self['tlvs'] += struct.pack('>I', hmac)
                        # Put hmac
                        self['tlvs'] += struct.pack('QQQQ', 0, 0, 0, 0)
                else:
                    raise TypeError(
                        'Family %s not supported for seg6 tunnel' % family
                    )
                # Finally encode as nla
                nla.encode(self)

            # Utility function to verify if hmac is present
            def has_hmac(self):
                # Useful during the decoding
                return self['flags'] & self.SR6_FLAG1_HMAC

            def decode(self):
                # Decode the data
                nla.decode(self)
                # Extract the encap mode
                self['mode'] = self.r_encapmodes.get(
                    self['encapmode'], "encap"
                )
                # Calculate offset of the segs
                offset = self.offset + 16
                # Point the addresses
                addresses = self.data[offset:]
                # Extract the number of segs
                n_segs = self['segments_left'] + 1
                # Init segs
                segs = []
                # Move 128 bit in each step
                for i in range(n_segs):
                    # Save the segment
                    segs.append(
                        inet_ntop(AF_INET6, addresses[i * 16 : i * 16 + 16])
                    )
                # Save segs
                self['segs'] = segs
                # Init tlvs
                self['tlvs'] = ''
                # If hmac is used
                if self.has_hmac():
                    # Point to the start of hmac
                    hmac = addresses[n_segs * 16 : n_segs * 16 + 40]
                    # Save tlvs section
                    self['tlvs'] = hexdump(hmac)
                    # Show also the hmac key
                    self['hmac'] = hexdump(hmac[4:8])

    class bpf_encap_info(nla):
        __slots__ = ()

        nla_map = (
            ('LWT_BPF_UNSPEC', 'none'),
            ('LWT_BPF_IN', 'bpf_obj'),
            ('LWT_BPF_OUT', 'bpf_obj'),
            ('LWT_BPF_XMIT', 'bpf_obj'),
            ('LWT_BPF_XMIT_HEADROOM', 'uint32'),
        )

        class bpf_obj(nla):
            __slots__ = ()

            nla_map = (
                ('LWT_BPF_PROG_UNSPEC', 'none'),
                ('LWT_BPF_PROG_FD', 'uint32'),
                ('LWT_BPF_PROG_NAME', 'asciiz'),
            )

    class seg6local_encap_info(nla):
        __slots__ = ()

        nla_map = (
            ('SEG6_LOCAL_UNSPEC', 'none'),
            ('SEG6_LOCAL_ACTION', 'action'),
            ('SEG6_LOCAL_SRH', 'ipv6_sr_hdr'),
            ('SEG6_LOCAL_TABLE', 'table'),
            ('SEG6_LOCAL_NH4', 'nh4'),
            ('SEG6_LOCAL_NH6', 'nh6'),
            ('SEG6_LOCAL_IIF', 'iif'),
            ('SEG6_LOCAL_OIF', 'oif'),
            ('SEG6_LOCAL_BPF', 'bpf_obj'),
            ('SEG6_LOCAL_VRFTABLE', 'vrf_table'),
        )

        class bpf_obj(nla):
            __slots__ = ()

            nla_map = (
                ('LWT_BPF_PROG_UNSPEC', 'none'),
                ('LWT_BPF_PROG_FD', 'uint32'),
                ('LWT_BPF_PROG_NAME', 'asciiz'),
            )

        class ipv6_sr_hdr(nla):
            __slots__ = ()

            fields = (
                ('nexthdr', 'B'),
                ('hdrlen', 'B'),
                ('type', 'B'),
                ('segments_left', 'B'),
                ('first_segment', 'B'),
                ('flags', 'B'),
                ('reserved', 'H'),
                ('segs', 's'),
                # Potentially several type-length-value
                ('tlvs', 's'),
            )

            # Corresponding values for seg6 encap modes
            SEG6_IPTUN_MODE_INLINE = 0
            SEG6_IPTUN_MODE_ENCAP = 1

            # Mapping string to nla value
            encapmodes = {
                "inline": SEG6_IPTUN_MODE_INLINE,
                "encap": SEG6_IPTUN_MODE_ENCAP,
            }

            # Reverse mapping: mapping nla value to string
            r_encapmodes = {v: k for k, v in encapmodes.items()}

            # Nla value for seg6 type
            SEG6_TYPE = 4

            # Flag value for hmac
            SR6_FLAG1_HMAC = 1 << 3

            # Tlv value for hmac
            SR6_TLV_HMAC = 5

            # Utility function to get the family from the msg
            def get_family(self):
                pointer = self
                while pointer.parent is not None:
                    pointer = pointer.parent
                return pointer.get('family', AF_UNSPEC)

            def encode(self):
                # Retrieve the family
                family = self.get_family()
                # Seg6 can be applied only to IPv6
                if family == AF_INET6:
                    # Get mode
                    mode = self['mode']
                    # Get segs
                    segs = self['segs']
                    # Get hmac
                    hmac = self.get('hmac', None)
                    # With "inline" mode there is not
                    # encap into an outer IPv6 header
                    if mode == "inline":
                        # Add :: to segs
                        segs.insert(0, "::")
                    # Add mode to value
                    self['encapmode'] = self.encapmodes.get(
                        mode, self.SEG6_IPTUN_MODE_ENCAP
                    )
                    # Calculate srlen
                    srhlen = 8 + 16 * len(segs)
                    # If we are using hmac we have a tlv as trailer data
                    if hmac:
                        # Since we can use sha1 or sha256
                        srhlen += 40
                    # Calculate and set hdrlen
                    self['hdrlen'] = (srhlen >> 3) - 1
                    # Add seg6 type
                    self['type'] = self.SEG6_TYPE
                    # Add segments left
                    self['segments_left'] = len(segs) - 1
                    # Add fitst segment
                    self['first_segment'] = len(segs) - 1
                    # If hmac is used we have to set the flags
                    if hmac:
                        # Add SR6_FLAG1_HMAC
                        self['flags'] |= self.SR6_FLAG1_HMAC
                    # Init segs
                    self['segs'] = b''
                    # Iterate over segments
                    for seg in segs:
                        # Convert to network byte order and add to value
                        self['segs'] += inet_pton(family, seg)
                    # Initialize tlvs
                    self['tlvs'] = b''
                    # If hmac is used we have to properly init tlvs
                    if hmac:
                        # Put type
                        self['tlvs'] += struct.pack('B', self.SR6_TLV_HMAC)
                        # Put length -> 40-2
                        self['tlvs'] += struct.pack('B', 38)
                        # Put reserved
                        self['tlvs'] += struct.pack('H', 0)
                        # Put hmac key
                        self['tlvs'] += struct.pack('>I', hmac)
                        # Put hmac
                        self['tlvs'] += struct.pack('QQQQ', 0, 0, 0, 0)
                else:
                    raise TypeError(
                        'Family %s not supported for seg6 tunnel' % family
                    )
                # Finally encode as nla
                nla.encode(self)

            # Utility function to verify if hmac is present
            def has_hmac(self):
                # Useful during the decoding
                return self['flags'] & self.SR6_FLAG1_HMAC

            def decode(self):
                # Decode the data
                nla.decode(self)
                # Extract the encap mode
                self['mode'] = self.r_encapmodes.get(
                    self['encapmode'], "encap"
                )
                # Calculate offset of the segs
                offset = self.offset + 16
                # Point the addresses
                addresses = self.data[offset:]
                # Extract the number of segs
                n_segs = self['segments_left'] + 1
                # Init segs
                segs = []
                # Move 128 bit in each step
                for i in range(n_segs):
                    # Save the segment
                    segs.append(
                        inet_ntop(AF_INET6, addresses[i * 16 : i * 16 + 16])
                    )
                # Save segs
                self['segs'] = segs
                # Init tlvs
                self['tlvs'] = ''
                # If hmac is used
                if self.has_hmac():
                    # Point to the start of hmac
                    hmac = addresses[n_segs * 16 : n_segs * 16 + 40]
                    # Save tlvs section
                    self['tlvs'] = hexdump(hmac)
                    # Show also the hmac key
                    self['hmac'] = hexdump(hmac[4:8])

        class table(nla):
            __slots__ = ()
            # Table ID
            fields = (('value', 'I'),)

        class action(nla):
            __slots__ = ()
            # Action
            fields = (('value', 'I'),)

            SEG6_LOCAL_ACTION_UNSPEC = 0
            SEG6_LOCAL_ACTION_END = 1
            SEG6_LOCAL_ACTION_END_X = 2
            SEG6_LOCAL_ACTION_END_T = 3
            SEG6_LOCAL_ACTION_END_DX2 = 4
            SEG6_LOCAL_ACTION_END_DX6 = 5
            SEG6_LOCAL_ACTION_END_DX4 = 6
            SEG6_LOCAL_ACTION_END_DT6 = 7
            SEG6_LOCAL_ACTION_END_DT4 = 8
            SEG6_LOCAL_ACTION_END_B6 = 9
            SEG6_LOCAL_ACTION_END_B6_ENCAP = 10
            SEG6_LOCAL_ACTION_END_BM = 11
            SEG6_LOCAL_ACTION_END_S = 12
            SEG6_LOCAL_ACTION_END_AS = 13
            SEG6_LOCAL_ACTION_END_AM = 14
            SEG6_LOCAL_ACTION_END_BPF = 15
            SEG6_LOCAL_ACTION_END_DT46 = 16

            actions = {
                'End': SEG6_LOCAL_ACTION_END,
                'End.X': SEG6_LOCAL_ACTION_END_X,
                'End.T': SEG6_LOCAL_ACTION_END_T,
                'End.DX2': SEG6_LOCAL_ACTION_END_DX2,
                'End.DX6': SEG6_LOCAL_ACTION_END_DX6,
                'End.DX4': SEG6_LOCAL_ACTION_END_DX4,
                'End.DT6': SEG6_LOCAL_ACTION_END_DT6,
                'End.DT4': SEG6_LOCAL_ACTION_END_DT4,
                'End.B6': SEG6_LOCAL_ACTION_END_B6,
                'End.B6.Encaps': SEG6_LOCAL_ACTION_END_B6_ENCAP,
                'End.BM': SEG6_LOCAL_ACTION_END_BM,
                'End.S': SEG6_LOCAL_ACTION_END_S,
                'End.AS': SEG6_LOCAL_ACTION_END_AS,
                'End.AM': SEG6_LOCAL_ACTION_END_AM,
                'End.BPF': SEG6_LOCAL_ACTION_END_BPF,
                'End.DT46': SEG6_LOCAL_ACTION_END_DT46,
            }

            def encode(self):
                # Get action type and convert string to value
                action = self['value']
                self['value'] = self.actions.get(
                    action, self.SEG6_LOCAL_ACTION_UNSPEC
                )
                # Convert action type to u32
                self['value'] = self['value'] & 0xFFFFFFFF
                # Finally encode as nla
                nla.encode(self)

        class iif(nla):
            __slots__ = ()

            # Index of the incoming interface
            fields = (('value', 'I'),)

        class oif(nla):
            __slots__ = ()

            # Index of the outcoming interface
            fields = (('value', 'I'),)

        class nh4(nla_string):
            __slots__ = ()

            # Nexthop of the IPv4 family

            def encode(self):
                # Convert to network byte order
                self['value'] = inet_pton(AF_INET, self['value'])
                # Finally encode as nla
                nla_string.encode(self)

            def decode(self):
                # Decode the data
                nla_string.decode(self)
                # Convert the packed IP address to its string representation
                self['value'] = inet_ntop(AF_INET, self['value'])

        class nh6(nla_string):
            __slots__ = ()

            # Nexthop of the IPv6 family

            def encode(self):
                # Convert to network byte order
                self['value'] = inet_pton(AF_INET6, self['value'])
                # Finally encode as nla
                nla_string.encode(self)

            def decode(self):
                # Decode the data
                nla_string.decode(self)
                # Convert the packed IP address to its string representation
                self['value'] = inet_ntop(AF_INET6, self['value'])

        class vrf_table(nla):
            __slots__ = ()

            # VRF Table ID
            fields = (('value', 'I'),)

    #
    # TODO: add here other lwtunnel types
    #
    encaps = {
        LWTUNNEL_ENCAP_MPLS: mpls_encap_info,
        LWTUNNEL_ENCAP_SEG6: seg6_encap_info,
        LWTUNNEL_ENCAP_BPF: bpf_encap_info,
        LWTUNNEL_ENCAP_SEG6_LOCAL: seg6local_encap_info,
    }

    class rta_mfc_stats(nla):
        __slots__ = ()

        fields = (
            ('mfcs_packets', 'uint64'),
            ('mfcs_bytes', 'uint64'),
            ('mfcs_wrong_if', 'uint64'),
        )

    class metrics(nla):
        __slots__ = ()

        prefix = 'RTAX_'
        nla_map = (
            ('RTAX_UNSPEC', 'none'),
            ('RTAX_LOCK', 'uint32'),
            ('RTAX_MTU', 'uint32'),
            ('RTAX_WINDOW', 'uint32'),
            ('RTAX_RTT', 'uint32'),
            ('RTAX_RTTVAR', 'uint32'),
            ('RTAX_SSTHRESH', 'uint32'),
            ('RTAX_CWND', 'uint32'),
            ('RTAX_ADVMSS', 'uint32'),
            ('RTAX_REORDERING', 'uint32'),
            ('RTAX_HOPLIMIT', 'uint32'),
            ('RTAX_INITCWND', 'uint32'),
            ('RTAX_FEATURES', 'uint32'),
            ('RTAX_RTO_MIN', 'uint32'),
            ('RTAX_INITRWND', 'uint32'),
            ('RTAX_QUICKACK', 'uint32'),
        )

    @staticmethod
    def get_nh(self, *argv, **kwarg):
        return nh

    class rtvia(nla_string):
        __slots__ = ()
        sql_type = 'TEXT'

        def encode(self):
            family = self.get('family', AF_UNSPEC)
            if family in (AF_INET, AF_INET6):
                addr = inet_pton(family, self['addr'])
            else:
                raise TypeError('Family %s not supported for RTA_VIA' % family)
            self['value'] = struct.pack('H', family) + addr
            nla_string.encode(self)

        def decode(self):
            nla_string.decode(self)
            family = struct.unpack('H', self['value'][:2])[0]
            addr = self['value'][2:]
            if addr:
                if (family == AF_INET and len(addr) == 4) or (
                    family == AF_INET6 and len(addr) == 16
                ):
                    addr = inet_ntop(family, addr)
                else:
                    addr = hexdump(addr)
            self.value = {'family': family, 'addr': addr}

    class cacheinfo(nla):
        __slots__ = ()

        fields = (
            ('rta_clntref', 'I'),
            ('rta_lastuse', 'I'),
            ('rta_expires', 'i'),
            ('rta_error', 'I'),
            ('rta_used', 'I'),
            ('rta_id', 'I'),
            ('rta_ts', 'I'),
            ('rta_tsage', 'I'),
        )


class rtmsg(rtmsg_base, nlmsg):
    __slots__ = ()

    def encode(self):
        if self.get('family') == AF_MPLS:
            # force fields
            self['dst_len'] = 20
            self['table'] = 254
            self['type'] = 1
            # assert NLA types
            for n in self.get('attrs', []):
                if n[0] not in (
                    'RTA_OIF',
                    'RTA_DST',
                    'RTA_VIA',
                    'RTA_NEWDST',
                    'RTA_MULTIPATH',
                ):
                    raise TypeError('Incorrect NLA type %s for AF_MPLS' % n[0])
        super(rtmsg_base, self).encode()


class nh(rtmsg_base, nla):
    __slots__ = ()

    is_nla = False

    sql_constraints = {}
    cell_header = (('length', 'H'),)
    fields = (('flags', 'B'), ('hops', 'B'), ('oif', 'i'))



from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import TC_H_INGRESS

parent = TC_H_INGRESS


def fix_request(request):
    if 'rate' in request:
        del request['rate']
    request['handle'] = 0xFFFF0000


class options(nla):
    fields = (('value', 'I'),)



from pyroute2.netlink import nla, nlmsg_atoms

CAN_CTRLMODE_NAMES = {
    'CAN_CTRLMODE_LOOPBACK': 0x01,
    'CAN_CTRLMODE_LISTENONLY': 0x02,
    'CAN_CTRLMODE_3_SAMPLES': 0x04,
    'CAN_CTRLMODE_ONE_SHOT': 0x08,
    'CAN_CTRLMODE_BERR_REPORTING': 0x10,
    'CAN_CTRLMODE_FD': 0x20,
    'CAN_CTRLMODE_PRESUME_ACK': 0x40,
    'CAN_CTRLMODE_FD_NON_ISO': 0x80,
    'CAN_CTRLMODE_CC_LEN8_DLC': 0x100,
    'CAN_CTRLMODE_TDC_AUTO': 0x200,
    'CAN_CTRLMODE_TDC_MANUAL': 0x400,
}

CAN_CTRLMODE_VALUES = {
    0x001: 'CAN_CTRLMODE_LOOPBACK',
    0x002: 'CAN_CTRLMODE_LISTENONLY',
    0x004: 'CAN_CTRLMODE_3_SAMPLES',
    0x008: 'CAN_CTRLMODE_ONE_SHOT',
    0x010: 'CAN_CTRLMODE_BERR_REPORTING',
    0x020: 'CAN_CTRLMODE_FD',
    0x040: 'CAN_CTRLMODE_PRESUME_ACK',
    0x080: 'CAN_CTRLMODE_FD_NON_ISO',
    0x100: 'CAN_CTRLMODE_CC_LEN8_DLC',
    0x200: 'CAN_CTRLMODE_TDC_AUTO',
    0x400: 'CAN_CTRLMODE_TDC_MANUAL',
}


class can(nla):
    prefix = 'IFLA_'
    nla_map = (
        ('IFLA_CAN_UNSPEC', 'none'),
        ('IFLA_CAN_BITTIMING', 'can_bittiming'),
        ('IFLA_CAN_BITTIMING_CONST', 'can_bittiming_const'),
        # NOTE:
        # This is actually a struct of one member, but that doesn't parse:
        ('IFLA_CAN_CLOCK', 'uint32'),
        ('IFLA_CAN_STATE', 'can_state'),
        ('IFLA_CAN_CTRLMODE', 'can_ctrlmode'),
        ('IFLA_CAN_RESTART_MS', 'uint32'),
        ('IFLA_CAN_RESTART', 'flag'),
        ('IFLA_CAN_BERR_COUNTER', 'can_berr_counter'),
        ('IFLA_CAN_DATA_BITTIMING', 'can_bittiming'),
        ('IFLA_CAN_DATA_BITTIMING_CONST', 'can_bittiming_const'),
        ('IFLA_CAN_TERMINATION', 'uint16'),
        ('IFLA_CAN_TERMINATION_CONST', 'array(uint16)'),
        ('IFLA_CAN_BITRATE_CONST', 'array(uint32)'),
        ('IFLA_CAN_DATA_BITRATE_CONST', 'array(uint32)'),
        ('IFLA_CAN_BITRATE_MAX', 'uint32'),
        ('IFLA_CAN_TDC', 'can_tdc'),
        ('IFLA_CAN_CTRLMODE_EXT', 'can_ctrlmode_ext'),
    )

    class can_bittiming(nla):
        fields = (
            ('bitrate', 'I'),
            ('sample_point', 'I'),
            ('tq', 'I'),
            ('prop_seg', 'I'),
            ('phase_seg1', 'I'),
            ('phase_seg2', 'I'),
            ('sjw', 'I'),
            ('brp', 'I'),
        )

    class can_bittiming_const(nla):
        fields = (
            ('name', '=16s'),
            ('tseg1_min', 'I'),
            ('tseg1_max', 'I'),
            ('tseg2_min', 'I'),
            ('tseg2_max', 'I'),
            ('sjw_max', 'I'),
            ('brp_min', 'I'),
            ('brp_max', 'I'),
            ('brp_inc', 'I'),
        )

    class can_state(nlmsg_atoms.uint32):
        value_map = {
            0: 'ERROR_ACTIVE',
            1: 'ERROR_WARNING',
            2: 'ERROR_PASSIVE',
            3: 'BUS_OFF',
            4: 'STOPPED',
            5: 'SLEEPING',
            6: 'MAX',
        }

    class can_ctrlmode(nla):
        fields = (('mask', 'I'), ('flags', 'I'))

        def decode(self):
            super(nla, self).decode()
            flags = self["flags"]
            for value, mode in CAN_CTRLMODE_VALUES.items():
                self[mode[len('CAN_CTRLMODE_') :].lower()] = (
                    "on" if flags & value else "off"
                )
            del self["flags"]
            del self["mask"]

        def encode(self):
            mask = 0
            flags = 0
            for mode, value in CAN_CTRLMODE_NAMES.items():
                m = mode[len('CAN_CTRLMODE_') :].lower()
                try:
                    v = self[m]
                except KeyError:
                    continue
                mask |= value
                if v == "on":
                    flags |= value
            self['mask'] = mask
            self['flags'] = flags
            return super(nla, self).encode()

    class can_berr_counter(nla):
        fields = (('txerr', 'H'), ('rxerr', 'H'))

    class can_tdc(nla):
        prefix = "IFLA_"
        nla_map = (
            ('IFLA_CAN_TDC_UNSPEC', 'none'),
            ('IFLA_CAN_TDC_TDCV_MIN', 'uint32'),
            ('IFLA_CAN_TDC_TDCV_MAX', 'uint32'),
            ('IFLA_CAN_TDC_TDCO_MIN', 'uint32'),
            ('IFLA_CAN_TDC_TDCO_MAX', 'uint32'),
            ('IFLA_CAN_TDC_TDCF_MIN', 'uint32'),
            ('IFLA_CAN_TDC_TDCF_MAX', 'uint32'),
            ('IFLA_CAN_TDC_TDCV', 'uint32'),
            ('IFLA_CAN_TDC_TDCO', 'uint32'),
            ('IFLA_CAN_TDC_TDCF', 'uint32'),
        )

    class can_ctrlmode_ext(nla):
        prefix = "IFLA_"
        nla_map = (
            ('IFLA_CAN_CTRLMODE_UNSPEC', 'none'),
            ('IFLA_CAN_CTRLMODE_SUPPORTED', 'can_ctrlmode_supported'),
        )

        class can_ctrlmode_supported(nlmsg_atoms.uint32):
            def decode(self):
                super(nlmsg_atoms.uint32, self).decode()
                for value, mode in CAN_CTRLMODE_VALUES.items():
                    self[mode[len('CAN_CTRLMODE_') :].lower()] = (
                        'yes' if value & self["value"] else 'no'
                    )
                del self["value"]



from pyroute2 import DL

dl = DL()
for q in dl.get_dump():
    print(
        '%s\t%s'
        % (
            q.get_attr('DEVLINK_ATTR_BUS_NAME'),
            q.get_attr('DEVLINK_ATTR_DEV_NAME'),
        )
    )
dl.close()



"""Helper functions to build dhcp client messages."""

from dataclasses import dataclass
from typing import Iterable, Literal, Optional

from pyroute2.dhcp import enums
from pyroute2.dhcp.dhcp4msg import dhcp4msg
from pyroute2.dhcp.fsm import State
from pyroute2.dhcp.leases import Lease
from pyroute2.dhcp.xids import Xid

Parameters = Iterable[enums.dhcp.Option]


@dataclass
class _DHCPMessage:
    '''A DHCP message with some extra info from other layers.'''

    dhcp: dhcp4msg
    eth_src: str
    eth_dst: str = 'ff:ff:ff:ff:ff:ff'
    ip_src: str = '0.0.0.0'
    ip_dst: str = '255.255.255.255'
    sport: int = 68
    dport: int = 67

    @property
    def message_type(self) -> enums.dhcp.MessageType:
        '''The DHCP message type (DISCOVER, REQUEST, ACK...)'''
        return self.dhcp['options']['message_type']

    @property
    def xid(self) -> Xid:
        return Xid(self.dhcp['xid'])


@dataclass
class SentDHCPMessage(_DHCPMessage):
    '''A DHCP message to be sent to a server or broadcast.'''

    eth_src: Optional[str] = None  # type: ignore[assignment]

    def __str__(self) -> str:
        type_name = self.message_type.name
        return (
            f'{type_name} to {self.eth_dst}/{self.ip_dst}:{self.dport} '
            f'(xid {self.xid})'
        )


class ReceivedDHCPMessage(_DHCPMessage):
    '''A DHCP message received by the client.'''

    def __str__(self) -> str:
        type_name = self.dhcp['options']['message_type'].name
        return (
            f'{type_name} from {self.eth_src}/{self.ip_src}:{self.sport} '
            f'(xid {self.xid})'
        )


def discover(parameter_list: Parameters) -> SentDHCPMessage:
    '''Make a broadcast DISCOVER message for the given parameters.'''
    # Default for SentDHCPMessage is broadcast which is what we want here
    return SentDHCPMessage(
        dhcp=dhcp4msg(
            {
                'op': enums.bootp.MessageType.BOOTREQUEST,
                'flags': enums.bootp.Flag.BROADCAST,
                'options': {
                    'message_type': enums.dhcp.MessageType.DISCOVER,
                    'parameter_list': list(parameter_list),
                },
            }
        )
    )


def request_for_offer(
    parameter_list: Parameters, offer: ReceivedDHCPMessage
) -> SentDHCPMessage:
    '''Make a REQUEST message for a given OFFER.

    Since we don't have an IP yet, the message is always broadcast.
    When requesting an offer in the Selecting state, the server_id DHCP option
    is always set as opposed to when a REQUEST is sent in other states.

    See RFC 2131 section 4.3.2.
    '''
    return SentDHCPMessage(
        dhcp=dhcp4msg(
            {
                'op': enums.bootp.MessageType.BOOTREQUEST,
                'flags': enums.bootp.Flag.BROADCAST,
                'options': {
                    'message_type': enums.dhcp.MessageType.REQUEST,
                    'requested_ip': offer.dhcp['yiaddr'],
                    'server_id': offer.dhcp['options']['server_id'],
                    'parameter_list': list(parameter_list),
                },
            }
        )
    )


def request_for_lease(
    parameter_list: Parameters,
    lease: Lease,
    state: Literal[State.RENEWING, State.REBINDING, State.REBOOTING],
) -> SentDHCPMessage:
    '''Make a REQUEST for an existing lease.

    This differs from REQUESTs in response to an OFFER in that the server_id
    option is never set.

    When rebooting, the message is broadcast, and the requested_ip option is
    set to the IP in the stored lease. The bootp client IP is left blank.

    When renewing, (i.e. T1 expires) the message is for the server that granted
    the lease. The lease's IP is expected to be assigned to the client's
    interface at this point.

    When rebinding (T2), the message is broadcast on the network.

    In both cases, the bootp client IP (ciaddr) is set to the lease's IP.

    See RFC 2131 section 4.3.6.
    '''
    dhcp_msg = dhcp4msg(
        {
            'op': enums.bootp.MessageType.BOOTREQUEST,
            'flags': enums.bootp.Flag.BROADCAST,
            'options': {
                'message_type': enums.dhcp.MessageType.REQUEST,
                'parameter_list': list(parameter_list),
            },
        }
    )
    if state == State.REBOOTING:
        dhcp_msg['options']['requested_ip'] = lease.ip
    else:
        dhcp_msg['ciaddr'] = lease.ip
        if state == State.RENEWING:
            # T1 timer expired, send a request directly to the known server
            dhcp_msg['flags'] = enums.bootp.Flag.UNICAST
            return SentDHCPMessage(
                dhcp=dhcp_msg,
                eth_dst=lease.server_mac,
                # default ip_dst to broadcast, it should not be necessary
                # but it'll avoid a crash if the server_id is missing
                ip_dst=lease.server_id or '255.255.255.255',
                ip_src=lease.ip,
            )
    # Reboot or rebind, broadcast the request
    return SentDHCPMessage(dhcp=dhcp_msg)


def release(lease: Lease) -> SentDHCPMessage:
    '''Make a RELEASE for an existing & active lease.'''
    # RELEASE messages have nearly no allowed options,
    # and the released IP address must be set in ciaddr
    return SentDHCPMessage(
        dhcp=dhcp4msg(
            {
                'op': enums.bootp.MessageType.BOOTREQUEST,
                'flags': enums.bootp.Flag.UNICAST,
                'ciaddr': lease.ip,
                'options': {
                    'message_type': enums.dhcp.MessageType.RELEASE,
                    'server_id': lease.server_id,
                },
            }
        ),
        # RELEASEs are unicast (see rfc section 4.4.4)
        eth_dst=lease.server_mac,
        # default ip_dst to broadcast, it should not be necessary
        # but it'll avoid a crash if the server_id is missing
        ip_dst=lease.server_id or '255.255.255.255',
        ip_src=lease.ip,
    )



import warnings

from pyroute2 import netns
from pyroute2.common import basestring
from pyroute2.netlink.rtnl.nsinfmsg import nsinfmsg
from pyroute2.requests.netns import NetNSFieldFilter

from ..objects import AsyncObject


async def load_nsinfmsg(schema, sources, target, event):
    #
    # check if there is corresponding source
    #
    netns_path = event.get_attr('NSINFO_PATH')
    if netns_path is None:
        schema.log.debug('ignore %s %s' % (target, event))
        return
    if schema.config['auto_netns']:
        warnings.warn(
            'automatic netns sourcing is being refactored', DeprecationWarning
        )
    await schema.load_netlink('netns', sources, target, event)


schema = nsinfmsg.sql_schema().unique_index('NSINFO_PATH')

init = {
    'specs': [['netns', schema]],
    'classes': [['netns', nsinfmsg]],
    'event_map': {nsinfmsg: [load_nsinfmsg]},
}


class NetNS(AsyncObject):
    table = 'netns'
    msg_class = nsinfmsg
    table_alias = 'n'
    api = 'netns'
    field_filter = NetNSFieldFilter

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = nsinfmsg
        self.event_map = {nsinfmsg: "load_rtnlmsg"}
        super(NetNS, self).__init__(*argv, **kwarg)

    @classmethod
    def spec_normalize(cls, spec):
        if isinstance(spec, basestring):
            spec = {'path': spec}
        path = netns._get_netnspath(spec['path'])
        # on Python3 _get_netnspath() returns bytes, not str, so
        # we have to decode it here in order to avoid issues with
        # cache keys and DB inserts
        if hasattr(path, 'decode'):
            path = path.decode('utf-8')
        spec['path'] = path
        return spec

    def __setitem__(self, key, value):
        if self.state == 'system':
            raise ValueError('attempt to change a readonly object')
        if key == 'path':
            value = netns._get_netnspath(value).decode('utf-8')
        return super(NetNS, self).__setitem__(key, value)



'''

Local RTNL
----------

Local RTNL source is a simple `IPRoute` instance. By default NDB
starts with one local RTNL source names `localhost`::

    >>> ndb = NDB()
    >>> ndb.sources.summary().format("json")
    [
        {
            "name": "localhost",
            "spec": "{'target': 'localhost', 'nlm_generator': 1}",
            "state": "running"
        },
        {
            "name": "localhost/nsmanager",
            "spec": "{'target': 'localhost/nsmanager'}",
            "state": "running"
        }
    ]
    >>> ndb.sources['localhost']
    [running] <IPRoute {'target: 'localhost', 'nlm_generator': 1}>

The `localhost` RTNL source starts an additional async cache thread.
The `nlm_generator` option means that instead of collections the
`IPRoute` object returns generators, so `IPRoute` responses will not
consume memory regardless of the RTNL objects number::

    >>> ndb.sources['localhost'].nl.link('dump')
    <generator object RTNL_API.filter_messages at 0x7f61a99a34a0>

See also: :ref:`iproute`

Network namespaces
------------------

There are two ways to connect additional sources to an NDB instance.
One is to specify sources when creating an NDB object::

    ndb = NDB(sources=[{'target': 'localhost'}, {'netns': 'test01'}])

Another way is to call `ndb.sources.add()` method::

    ndb.sources.add(netns='test01')

This syntax: `{target': 'localhost'}` and `{'netns': 'test01'}` is the
short form. The full form would be::

    {'target': 'localhost', # the label for the DB
     'kind': 'local',       # use IPRoute class to start the source
     'nlm_generator': 1}    #

    {'target': 'test01',    # the label
     'kind': 'netns',       # use NetNS class
     'netns': 'test01'}     #

See also: :ref:`netns`
'''

import errno
import importlib
import queue
import socket
import struct
import sys
import threading
import time
import uuid

from pyroute2.common import basestring
from pyroute2.iproute import AsyncIPRoute
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg

from .events import ShutdownException, State
from .messages import cmsg_event
from .objects import RTNL_Object

if sys.platform.startswith('linux'):
    from pyroute2 import netns
    from pyroute2.iproute.linux import NetNS
    from pyroute2.netns.manager import NetNSManager
else:
    NetNS = None
    NetNSManager = None

SOURCE_FAIL_PAUSE = 1
SOURCE_MAX_ERRORS = 3


class SourceProxy(object):
    def __init__(self, ndb, target):
        self.ndb = ndb
        self.events = queue.Queue()
        self.target = target

    def api(self, name, *argv, **kwarg):
        call_id = str(uuid.uuid4().hex)
        self.ndb._call_registry[call_id] = event = threading.Event()
        event.clear()
        (
            self.ndb.messenger.emit(
                {
                    'type': 'api',
                    'target': self.target,
                    'call_id': call_id,
                    'name': name,
                    'argv': argv,
                    'kwarg': kwarg,
                }
            )
        )

        event.wait()
        response = self.ndb._call_registry.pop(call_id)
        if 'return' in response:
            return response['return']
        elif 'exception' in response:
            raise response['exception']


class Source(dict):
    '''
    The RNTL source. The source that is used to init the object
    must comply to IPRoute API, must support the async_cache. If
    the source starts additional threads, they must be joined
    in the source.close()
    '''

    table_alias = 'src'
    dump_header = None
    summary_header = None
    view = None
    table = 'sources'
    vmap = {'local': AsyncIPRoute, 'netns': NetNS, 'nsmanager': NetNSManager}

    def __init__(self, ndb, **spec):
        self.th = None
        self.task = None
        self.nl = None
        self.ndb = ndb
        self.evq = self.ndb.task_manager.event_queue
        # the target id -- just in case
        self.target = spec['target']
        self.kind = spec.pop('kind', 'local')
        self.max_errors = spec.pop('max_errors', SOURCE_MAX_ERRORS)
        self.event = spec.pop('event')
        # RTNL API
        self.nl_kwarg = spec
        self.nl_kwarg['nlm_echo'] = True
        #
        if self.ndb.messenger is not None:
            self.ndb.messenger.targets.add(self.target)
        #
        self.errors_counter = 0
        self.exception = None
        self.shutdown = threading.Event()
        self.started = threading.Event()
        self.lock = threading.RLock()
        self.shutdown_lock = threading.RLock()
        self.started.clear()
        self.log = ndb.log.channel('sources.%s' % self.target)
        self.state = State(log=self.log, wait_list=['running'])
        self.state.set('init')
        self.ndb.schema.add_nl_source(self.target, self.kind, spec)
        self.load_sql()

    @classmethod
    def _count(cls, view):
        return view.ndb.schema.fetchone(f'SELECT count(*) FROM {view.table}')

    @property
    def must_restart(self):
        if self.max_errors < 0 or self.errors_counter <= self.max_errors:
            return True
        return False

    @property
    def bind_arguments(self):
        return dict(
            filter(
                lambda x: x[1] is not None,
                (
                    ('async_cache', True),
                    ('clone_socket', True),
                    ('groups', self.nl_kwarg.get('groups')),
                ),
            )
        )

    async def set_ready(self):
        try:
            if self.event is not None:
                await self.evq.put(cmsg_event(self.target, self.event))
        except ShutdownException:
            self.state.set('stop')
            return False
        return True

    @classmethod
    def defaults(cls, spec):
        ret = dict(spec)
        defaults = {}
        if 'hostname' in spec:
            defaults['kind'] = 'remote'
            defaults['protocol'] = 'ssh'
            defaults['target'] = spec['hostname']
        if 'netns' in spec:
            defaults['kind'] = 'netns'
            defaults['target'] = spec['netns']
            ret['netns'] = netns._get_netnspath(spec['netns'])
        for key in defaults:
            if key not in ret:
                ret[key] = defaults[key]
        return ret

    def __repr__(self):
        return '[%s] <%s>' % (self.state.get(), self.nl_kwarg)

    @classmethod
    def nla2name(cls, name):
        return name

    @classmethod
    def name2nla(cls, name):
        return name

    @classmethod
    def summary(cls, view):
        yield ('state', 'name', 'spec')
        for key in view.keys():
            yield (view[key].state.get(), key, '%s' % (view[key].nl_kwarg,))

    @classmethod
    def dump(cls, view):
        return cls.summary(view)

    @classmethod
    def compare_record(self, left, right):
        # specific compare
        if isinstance(right, basestring):
            return right == left['name']

    def get_prime(self, name):
        return self.vmap.get(self.kind, None) or getattr(
            importlib.import_module('pyroute2'), self.kind
        )

    async def api(self, name, *argv, **kwarg):
        for _ in range(100):  # FIXME make a constant
            with self.lock:
                try:
                    self.log.debug(f'source api run {name} {argv} {kwarg}')
                    result = await getattr(self.nl, name)(*argv, **kwarg)
                    if isinstance(result, list):
                        for msg in result:
                            await self.evq.put(msg)
                    return result
                except (
                    NetlinkError,
                    AttributeError,
                    ValueError,
                    KeyError,
                    TypeError,
                    socket.error,
                    struct.error,
                ):
                    raise
                except Exception as e:
                    # probably the source is restarting
                    self.errors_counter += 1
                    self.log.debug(f'source api error: <{e}>')
                    time.sleep(1)
        raise RuntimeError('api call failed')

    async def fake_zero_if(self):
        url = 'https://github.com/svinota/pyroute2/issues/737'
        zero_if = ifinfmsg()
        zero_if['index'] = 0
        zero_if['state'] = 'up'
        zero_if['flags'] = 1
        zero_if['family'] = 0
        zero_if['header']['flags'] = 2
        zero_if['header']['type'] = 16
        zero_if['header']['target'] = self.target
        zero_if['event'] = 'RTM_NEWLINK'
        zero_if['attrs'] = [
            ('IFLA_IFNAME', url),
            ('IFLA_ADDRESS', '00:00:00:00:00:00'),
        ]
        zero_if.encode()
        await self.evq.put(zero_if)

    async def receiver(self):
        #
        # The source thread routine -- get events from the
        # channel and forward them into the common event queue
        #
        # The routine exists on an event with error code == 104
        #
        if self.nl is not None:
            try:
                self.nl.close(code=0)
            except Exception as e:
                self.log.warning('source restart: %s' % e)
        try:
            self.state.set('connecting')
            spec = {}
            spec.update(self.nl_kwarg)
            self.nl = AsyncIPRoute(**spec)
            self.state.set('loading')
            #
            await self.nl.setup_endpoint()
            await self.nl.bind(**self.bind_arguments)
            self.log.debug(f'source fd {self.nl.fileno()}')
            #
            # Initial load -- enqueue the data
            #
            self.ndb.schema.flush(self.target)
            if self.kind in ('local', 'netns', 'remote'):
                await self.fake_zero_if()
            async for x in await self.nl.dump():
                await self.evq.put(x)
            self.state.set('running')
        finally:
            await self.set_ready()

        while self.state.get() not in ('stop', 'restart'):
            async for msg in self.nl.get():
                error = msg['header']['error']
                if error:
                    raise error
                await self.evq.put(msg)

    def close(self, code=errno.ECONNRESET, sync=True):
        with self.shutdown_lock:
            if self.shutdown.is_set():
                self.log.debug('already stopped')
                return
            self.log.debug('source shutdown')
            self.shutdown.set()
            if self.nl is not None:
                try:
                    self.nl.close(code=code)
                except Exception as e:
                    self.log.error('source close: %s' % e)
        if sync:
            if self.th is not None:
                self.th.join()
                self.th = None
            else:
                self.log.debug('receiver thread missing')

    def restart(self, reason='unknown'):
        with self.lock:
            with self.shutdown_lock:
                self.log.debug('restarting the source, reason <%s>' % (reason))
                self.started.clear()
                try:
                    self.close()
                    if self.th:
                        self.th.join()
                    self.shutdown.clear()
                    self.start()
                finally:
                    pass
        self.started.wait()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def load_sql(self):
        #
        spec = self.ndb.schema.fetchone(
            'SELECT * FROM sources WHERE f_target = ?', (self.target,)
        )
        self['target'], self['kind'] = spec
        for spec in self.ndb.schema.fetch(
            'SELECT * FROM sources_options WHERE f_target = ?', (self.target,)
        ):
            f_target, f_name, f_type, f_value = spec
            self[f_name] = int(f_value) if f_type == 'int' else f_value


class SyncSource(RTNL_Object):

    @property
    def nl(self):
        return self.asyncore.nl

    def api(self, name, *argv, **kwarg):
        return self._main_sync_call(self.asyncore.api, name, *argv, **kwarg)

    def set(self, key, value):
        if key == 'state':
            self.asyncore.ndb.task_manager.task_map[
                self.asyncore.task_id
            ].state.set(value)
            return
        raise RuntimeError('unknown property')

    def restart(self, reason='no reason'):
        self.asyncore.event.clear()
        self.close(next_state='running')
        self._main_async_call(self.asyncore.event.wait)

    def close(self, code=errno.ECONNRESET, sync=None, next_state='stopped'):
        self.set('state', next_state)
        self._main_sync_call(
            self.asyncore.ndb.schema.flush, self.asyncore.target
        )
        self._main_sync_call(self.asyncore.nl.close, code)



'''
IPVS -- IP Virtual Server
-------------------------

IPVS configuration is done via generic netlink protocol.
At the low level one can use it with a GenericNetlinkSocket,
binding it to "IPVS" generic netlink family.

But for the convenience the library provides utility classes:

    * IPVS -- a socket class to access the API
    * IPVSService -- a class to define IPVS service records
    * IPVSDest -- a class to define real server records

Dump all the records::

    from pyroute2 import IPVS, IPVSDest, IPVSService

    # run the socket
    ipvs = IPVS()

    # iterate all the IPVS services
    for s in ipvs.service("dump"):

        # create a utility object from a netlink message
        service = IPVSService.from_message(s)
        print("Service: ", service)

        # iterate all the real servers for this service
        for d in ipvs.dest("dump", service=service):

            # create and print a utility object
            dest = IPVSDest.from_message(d)
            print("  Real server: ", dest)

Create a service and a real server record::

    from socket import IPPROTO_TCP
    from pyroute2 import IPVS, IPVSDest, IPVSService

    ipvs = IPVS()

    service = IPVSService(addr="192.168.122.1", port=80, protocol=IPPROTO_TCP)
    real_server = IPVSDest(addr="10.0.2.20", port=80)

    ipvs.service("add", service=service)
    ipvs.dest("add", service=service, dest=real_server)

Delete a service::

    from pyroute2 import IPVS, IPVSService

    ipvs = IPVS()
    ipvs.service("del",
        service=IPVSService(
            addr="192.168.122.1",
            port=80,
            protocol=IPPROTO_TCP
        )
    )

'''

from socket import AF_INET

from pyroute2.common import get_address_family
from pyroute2.netlink.generic import ipvs
from pyroute2.netlink.nlsocket import NetlinkRequest
from pyroute2.requests.common import NLAKeyTransform
from pyroute2.requests.main import RequestProcessor


class ServiceFieldFilter(NLAKeyTransform):
    _nla_prefix = 'IPVS_SVC_ATTR_'

    def set_addr(self, context, value):
        ret = {"addr": value}
        if "af" in context.keys():
            family = context["af"]
        else:
            family = ret["af"] = get_address_family(value)
        if family == AF_INET and "netmask" not in context.keys():
            ret["netmask"] = "255.255.255.255"
        return ret


class DestFieldFilter(NLAKeyTransform):
    _nla_prefix = 'IPVS_DEST_ATTR_'

    def set_addr(self, context, value):
        ret = {"addr": value}
        if "addr_family" not in context.keys():
            ret["addr_family"] = get_address_family(value)
        return ret


class NLAFilter(RequestProcessor):
    msg = None
    keys = tuple()
    field_filters = None
    nla = None
    default_values = {}

    def __init__(self, **kwarg):
        dict.update(self, self.default_values)
        # save field filters
        flt = self.field_filters
        # init resets the filters, not fixed yet
        super().__init__(prime=kwarg)
        # restore filters
        for f in flt:
            self.add_filter(f)

    @classmethod
    def from_message(cls, msg):
        obj = cls()
        for key, value in msg.get(cls.nla)["attrs"]:
            obj[key] = value
        obj.pop("stats", None)
        obj.pop("stats64", None)
        return obj

    def dump_nla(self, items=None):
        if items is None:
            items = self.items()
        self.update(self)
        self.finalize()
        return {
            "attrs": list(
                map(lambda x: (self.msg.name2nla(x[0]), x[1]), items)
            )
        }

    def dump_key(self):
        return self.dump_nla(
            items=filter(lambda x: x[0] in self.key_fields, self.items())
        )


class IPVSService(NLAFilter):
    field_filters = [ServiceFieldFilter()]
    msg = ipvs.ipvsmsg.service
    key_fields = ("af", "protocol", "addr", "port")
    nla = "IPVS_CMD_ATTR_SERVICE"
    default_values = {
        "timeout": 0,
        "sched_name": "wlc",
        "flags": {"flags": 0, "mask": 0xFFFF},
    }


class IPVSDest(NLAFilter):
    field_filters = [DestFieldFilter()]
    msg = ipvs.ipvsmsg.dest
    nla = "IPVS_CMD_ATTR_DEST"
    default_values = {
        "fwd_method": 3,
        "weight": 1,
        "tun_type": 0,
        "tun_port": 0,
        "tun_flags": 0,
        "u_thresh": 0,
        "l_thresh": 0,
    }


class IPVS(ipvs.IPVSSocket):

    def service(self, command, service=None):
        command_map = {
            "add": (ipvs.IPVS_CMD_NEW_SERVICE, "create"),
            "set": (ipvs.IPVS_CMD_SET_SERVICE, "change"),
            "update": (ipvs.IPVS_CMD_DEL_SERVICE, "change"),
            "del": (ipvs.IPVS_CMD_DEL_SERVICE, "req"),
            "get": (ipvs.IPVS_CMD_GET_SERVICE, "get"),
            "dump": (ipvs.IPVS_CMD_GET_SERVICE, "dump"),
        }
        cmd, flags = NetlinkRequest.calculate_request_type(
            command, command_map
        )
        msg = ipvs.ipvsmsg()
        msg["cmd"] = cmd
        msg["version"] = ipvs.GENL_VERSION
        if service is not None:
            msg["attrs"] = [("IPVS_CMD_ATTR_SERVICE", service.dump_nla())]
        return self.nlm_request(msg, msg_type=self.prid, msg_flags=flags)

    def dest(self, command, service, dest=None):
        command_map = {
            "add": (ipvs.IPVS_CMD_NEW_DEST, "create"),
            "set": (ipvs.IPVS_CMD_SET_DEST, "change"),
            "update": (ipvs.IPVS_CMD_DEL_DEST, "change"),
            "del": (ipvs.IPVS_CMD_DEL_DEST, "req"),
            "get": (ipvs.IPVS_CMD_GET_DEST, "get"),
            "dump": (ipvs.IPVS_CMD_GET_DEST, "dump"),
        }
        cmd, flags = self.make_request_type(command, command_map)
        msg = ipvs.ipvsmsg()
        msg["cmd"] = cmd
        msg["version"] = 0x1
        msg["attrs"] = [("IPVS_CMD_ATTR_SERVICE", service.dump_key())]
        if dest is not None:
            msg["attrs"].append(("IPVS_CMD_ATTR_DEST", dest.dump_nla()))
        return self.nlm_request(msg, msg_type=self.prid, msg_flags=flags)



import errno
import os
import struct
from collections.abc import Callable
from typing import Optional
from unittest import mock

from pyroute2 import config
from pyroute2.common import msg_done
from pyroute2.iproute.ipmock import IPEngine
from pyroute2.netlink import NETLINK_ROUTE, NLM_F_REQUEST, nlmsg, rtnl
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.nlsocket import (
    AsyncNetlinkSocket,
    ChaoticNetlinkSocket,
    NetlinkRequest,
    NetlinkSocket,
)
from pyroute2.netlink.rtnl.ifinfmsg.tuntap import manage_tun, manage_tuntap
from pyroute2.netlink.rtnl.marshal import MarshalRtnl
from pyroute2.netlink.rtnl.probe_msg import proxy_newprobe
from pyroute2.netns import setns
from pyroute2.process import ChildProcess, ChildProcessReturnValue


def _run_in_netns(
    netns: str, target: Callable[[nlmsg], bytes], msg: nlmsg
) -> ChildProcessReturnValue:
    setns(netns)
    return ChildProcessReturnValue(target(msg), [])


class IPRouteProxy:
    route: dict[
        int,
        tuple[tuple[Callable[[nlmsg], bool], Callable[[nlmsg], bytes]], ...],
    ] = {
        rtnl.RTM_NEWLINK: (
            (lambda x: x.get(('linkinfo', 'kind')) == 'tuntap', manage_tuntap),
            (lambda x: x.get(('linkinfo', 'kind')) == 'tun', manage_tun),
        ),
        rtnl.RTM_NEWPROBE: ((lambda x: True, proxy_newprobe),),
    }

    def __init__(self, netns: Optional[str] = None):
        self.netns = netns

    def handle(self, msg: nlmsg) -> bytes:
        ret: bytes = b''
        key = msg['header']['type']
        if key not in self.route:
            return ret
        for predicate, target in self.route[key]:
            if predicate(msg):
                try:
                    if self.netns is not None:
                        with ChildProcess(
                            target=_run_in_netns,
                            args=[self.netns, target, msg],
                        ) as proc:
                            ret = proc.get_data(timeout=4)
                    else:
                        ret = target(msg)
                    return ret or msg_done(msg)
                except Exception as e:
                    # errmsg
                    if isinstance(e, (OSError, IOError)):
                        code = e.errno or errno.ENODATA
                    elif isinstance(e, NetlinkError):
                        code = e.code
                    else:
                        code = errno.ECOMM
                    newmsg = struct.pack('HH', 2, 0)
                    newmsg += msg.data[8:16]
                    newmsg += struct.pack('I', code)
                    newmsg += msg.data
                    newmsg = struct.pack('I', len(newmsg) + 4) + newmsg
                    return newmsg
        return b''


class AsyncIPRSocket(AsyncNetlinkSocket):
    '''A low-level class to provide RTNL socket.

    This is a low-level class designed to provide an RTNL
    asyncio-controlled socket. It does not include high-level
    methods like those found in AsyncIPRoute. Instead, it provides
    only common netlink methods such as `get()` and `put()`. For
    more details, refer to the `AsyncNetlinkSocket` documentation.

    .. testcode::
        :hide:

        from pyroute2 import AsyncIPRSocket

        iprsock = AsyncIPRSocket()
        assert callable(iprsock.get)
        assert callable(iprsock.put)
        assert callable(iprsock.nlm_request)
        assert callable(iprsock.bind)

    Since the underlying socket is controlled by asyncio, it is
    not possible to use it in poll/select loops. If you want
    such API, consider using synchronous `IPRSocket`.

    .. warning::

        Your code must process incoming messages quickly enough to
        prevent the RCVBUF from overflowing. If the RCVBUF overflows,
        all subsequent socket operations will raise an OSError:

    .. code::

        >>> [ x async for x in ipr.get() ]
        Traceback (most recent call last):
          File ".../python3.13/futures/_base.py", line 456, in result
            return self.__get_result()
                   ~~~~~~~~~~~~~~~~~^^
          ...
        OSError: [Errno 105] No buffer space available


    If this exception occurs, the only solution is to close the
    socket and create a new one.

    This class does not handle protocol-level error propagation; it
    only provides socket-level error handling. It is the user's
    responsibility to catch and manage protocol-level errors:

    .. testsetup:: as0

        from pyroute2.netlink import nlmsgerr, NLMSG_ERROR
        msg = nlmsgerr()
        msg['header']['type'] = NLMSG_ERROR
        msg['error'] = 42
        msg.reset()
        msg.encode()
        msg.decode()

    .. testcode:: as0

        if msg.get(('header', 'type')) == NLMSG_ERROR:
            # prints error code and the request that
            # triggered the error
            print(
                msg.get('error'),
                msg.get('msg'),
            )

    .. testoutput:: as0
        :hide:

        42 None


    '''

    def __init__(
        self,
        port=None,
        pid=None,
        fileno=None,
        sndbuf=1048576,
        rcvbuf=1048576,
        rcvsize=16384,
        all_ns=False,
        async_qsize=None,
        nlm_generator=None,
        target='localhost',
        ext_ack=False,
        strict_check=False,
        groups=rtnl.RTMGRP_DEFAULTS,
        nlm_echo=False,
        netns=None,
        netns_path=None,
        flags=os.O_CREAT,
        libc=None,
        use_socket=None,
        use_event_loop=None,
        telemetry=None,
    ):
        if config.mock_netlink:
            use_socket = IPEngine()
            if netns is not None:
                use_socket.netns = netns
            use_socket.flags = flags
            use_socket.initdb()
        self.marshal = MarshalRtnl()
        super().__init__(
            family=NETLINK_ROUTE,
            port=port,
            pid=pid,
            fileno=fileno,
            sndbuf=sndbuf,
            rcvbuf=rcvbuf,
            rcvsize=rcvsize,
            all_ns=all_ns,
            async_qsize=async_qsize,
            nlm_generator=nlm_generator,
            target=target,
            ext_ack=ext_ack,
            strict_check=strict_check,
            groups=groups,
            nlm_echo=nlm_echo,
            use_socket=use_socket,
            netns=netns,
            flags=flags,
            libc=libc,
            use_event_loop=use_event_loop,
            telemetry=telemetry,
        )
        if not config.mock_netlink:
            self.request_proxy = IPRouteProxy(netns)
        self.status['netns_path'] = netns_path or config.netns_path

    async def bind(self, groups=None, **kwarg):
        return await super().bind(
            groups if groups is not None else self.status['groups'], **kwarg
        )


class NotLocal:
    event_loop = None
    msg_queue = mock.Mock()


class IPRSocket(NetlinkSocket):
    '''Synchronous select-compatible netlink socket.

    `IPRSocket` is the synchronous counterpart to `AsyncIPRSocket`.
    A key feature of `IPRSocket` is that the underlying netlink
    socket operates out of asyncio control, allowing it to be
    used in poll/select loops.

    .. warning::

        Your code must process incoming messages quickly enough to
        prevent the RCVBUF from overflowing. If the RCVBUF overflows,
        all subsequent socket operations will raise an OSError:

    .. code::

        >>> iprsock.get()
        Traceback (most recent call last):
          File "<python-input-12>", line 1, in <module>
            iprsock.get()
            ~~~~~~~~^^
          File ".../pyroute2/netlink/rtnl/iprsocket.py", line 276, in get
            data = self.socket.recv(16384)
        OSError: [Errno 105] No buffer space available
        >>>

    If this exception occurs, the only solution is to close the
    socket and create a new one.

    Some usage examples:

    .. testcode::

        import select

        from pyroute2 import IPRSocket
        from pyroute2.netlink import NLM_F_DUMP, NLM_F_REQUEST
        from pyroute2.netlink.rtnl import RTM_GETLINK
        from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg

        with IPRSocket() as iprsock:
            iprsock.put(
                ifinfmsg(),
                msg_type=RTM_GETLINK,
                msg_flags=NLM_F_REQUEST | NLM_F_DUMP
            )

            ret = []

            while True:
                rl, wl, xl = select.select([iprsock], [], [], 0)
                if not len(rl):
                    break
                ret.extend(iprsock.get())

            for link in ret:
                if link.get('event') == 'RTM_NEWLINK':
                    print(
                        link.get('ifname'),
                        link.get('state'),
                        link.get('address'),
                    )

    .. testoutput::

        lo up 00:00:00:00:00:00
        eth0 up 52:54:00:72:58:b2

    Threadless RT netlink monitoring with blocking I/O calls:

        >>> from pyroute2 import IPRSocket
        >>> from pprint import pprint
        >>> s = IPRSocket()
        >>> s.bind()
        >>> pprint(s.get())
        [{'attrs': [('RTA_TABLE', 254),
                    ('RTA_OIF', 2),
                    ('RTA_GATEWAY', '192.168.122.1')],
          'dst_len': 0,
          'event': 'RTM_NEWROUTE',
          'family': 2,
          'flags': 0,
          'header': {'error': None,
                     'flags': 2,
                     'length': 52,
                     'pid': 325359,
                     'sequence_number': 255,
                     'type': 24},
          'proto': 2,
          'scope': 0,
          'src_len': 0,
          'table': 254,
          'tos': 0,
          'type': 2}]
        >>>

    Like `AsyncIPRSocket`, it does not perform response reassembly,
    protocol-level error propagation, or packet buffering.
    '''

    def __init__(
        self,
        port=None,
        pid=None,
        fileno=None,
        sndbuf=1048576,
        rcvbuf=1048576,
        rcvsize=16384,
        all_ns=False,
        async_qsize=None,
        nlm_generator=None,
        target='localhost',
        ext_ack=False,
        strict_check=False,
        groups=rtnl.RTMGRP_DEFAULTS,
        nlm_echo=False,
        netns=None,
        netns_path=None,
        flags=os.O_CREAT,
        libc=None,
        use_socket=None,
        use_event_loop=None,
    ):
        self.asyncore = AsyncIPRSocket(
            port=port,
            pid=pid,
            fileno=fileno,
            sndbuf=sndbuf,
            rcvbuf=rcvbuf,
            rcvsize=rcvsize,
            all_ns=all_ns,
            target=target,
            ext_ack=ext_ack,
            strict_check=strict_check,
            groups=groups,
            nlm_echo=nlm_echo,
            netns=netns,
            netns_path=netns_path,
            flags=flags,
            libc=libc,
            use_socket=use_socket,
            use_event_loop=use_event_loop,
        )
        self.asyncore.local = NotLocal()

    @property
    def addr_pool(self):
        return self.asyncore.addr_pool

    @property
    def socket(self):
        return self.asyncore.socket

    @property
    def fileno(self):
        return self.asyncore.local.socket.fileno

    def bind(self, groups=None, pid=None, **kwarg):
        self.asyncore._sync_bind(groups, pid, **kwarg)

    def put(
        self,
        msg,
        msg_type,
        msg_flags=NLM_F_REQUEST,
        addr=(0, 0),
        msg_seq=0,
        msg_pid=None,
    ):
        if msg is None:
            msg_class = self.marshal.msg_map[msg_type]
            msg = msg_class()

        request = NetlinkRequest(
            self,
            msg,
            msg_type=msg_type,
            msg_flags=msg_flags,
            msg_seq=msg_seq,
            msg_pid=msg_pid,
        )
        request.msg.encode()
        return request.sock.send(request.msg.data)

    def get(self, msg_seq=0, terminate=None, callback=None, noraise=False):
        data = self.socket.recv(16384)
        return [x for x in self.marshal.parse(data)]


class ChaoticIPRSocket(AsyncIPRSocket, ChaoticNetlinkSocket):
    pass



##
#
# This module contains all the public symbols from the library.
#

##
#
# Version
#
try:
    from pyroute2.config.version import __version__
except ImportError:
    __version__ = 'unknown'
import sys

##
#
# Logging setup
#
# See the history:
#  * https://github.com/svinota/pyroute2/issues/246
#  * https://github.com/svinota/pyroute2/issues/255
#  * https://github.com/svinota/pyroute2/issues/270
#  * https://github.com/svinota/pyroute2/issues/573
#  * https://github.com/svinota/pyroute2/issues/601
#
from pyroute2.config import log
from pyroute2.conntrack import AsyncConntrack, Conntrack, ConntrackEntry
from pyroute2.devlink import DL, AsyncDL
from pyroute2.ethtool.ethtool import Ethtool
from pyroute2.ipdb import IPDB, CommitException, CreateException
from pyroute2.iproute import (
    AsyncIPRoute,
    ChaoticIPRoute,
    IPBatch,
    IPRoute,
    NetNS,
    RawIPRoute,
)
from pyroute2.ipset import IPSet
from pyroute2.ipvs import IPVS, IPVSDest, IPVSService
from pyroute2.iwutil import IW, AsyncIW
from pyroute2.ndb.main import NDB
from pyroute2.netlink.connector.cn_proc import ProcEventSocket
from pyroute2.netlink.devlink import AsyncDevlinkSocket, DevlinkSocket
from pyroute2.netlink.diag import DiagSocket, ss2
from pyroute2.netlink.event import AsyncEventSocket, EventSocket
from pyroute2.netlink.event.acpi_event import (
    AcpiEventSocket,
    AsyncAcpiEventSocket,
)
from pyroute2.netlink.event.dquot import AsyncDQuotSocket, DQuotSocket
from pyroute2.netlink.event.thermal import (
    AsyncThermalEventSocket,
    ThermalEventSocket,
)
from pyroute2.netlink.exceptions import (
    ChaoticException,
    NetlinkDecodeError,
    NetlinkDumpInterrupted,
    NetlinkError,
)
from pyroute2.netlink.generic import (
    AsyncGenericNetlinkSocket,
    GenericNetlinkSocket,
)
from pyroute2.netlink.generic.ethtool import AsyncNlEthtool, NlEthtool
from pyroute2.netlink.generic.ipvs import AsyncIPVSSocket, IPVSSocket
from pyroute2.netlink.generic.l2tp import AsyncL2tp, L2tp
from pyroute2.netlink.generic.mptcp import MPTCP, AsyncMPTCP
from pyroute2.netlink.generic.wireguard import AsyncWireGuard, WireGuard
from pyroute2.netlink.ipq import IPQSocket
from pyroute2.netlink.nfnetlink.nfctsocket import AsyncNFCTSocket, NFCTSocket
from pyroute2.netlink.nfnetlink.nftsocket import AsyncNFTSocket, NFTSocket
from pyroute2.netlink.nl80211 import NL80211, AsyncNL80211
from pyroute2.netlink.rtnl.iprsocket import AsyncIPRSocket, IPRSocket
from pyroute2.netlink.taskstats import AsyncTaskStats, TaskStats
from pyroute2.netlink.uevent import UeventSocket
from pyroute2.nslink.nspopen import NSPopen
from pyroute2.plan9.client import Plan9ClientSocket
from pyroute2.plan9.server import Plan9ServerSocket
from pyroute2.wiset import WiSet

##
#
# Windows platform specific: socket module monkey patching
#
# To use the library on Windows, run::
#   pip install win-inet-pton
#
if sys.platform.startswith('win'):  # noqa: E402
    import win_inet_pton  # noqa: F401


modules = [
    AcpiEventSocket,
    AsyncAcpiEventSocket,
    AsyncConntrack,
    AsyncDL,
    AsyncDQuotSocket,
    AsyncDevlinkSocket,
    AsyncEventSocket,
    AsyncGenericNetlinkSocket,
    AsyncIPRSocket,
    AsyncIPRoute,
    AsyncIPVSSocket,
    AsyncIW,
    AsyncL2tp,
    AsyncMPTCP,
    AsyncNFCTSocket,
    AsyncNFTSocket,
    AsyncNL80211,
    AsyncNlEthtool,
    AsyncTaskStats,
    AsyncThermalEventSocket,
    AsyncWireGuard,
    ChaoticException,
    ChaoticIPRoute,
    CommitException,
    Conntrack,
    ConntrackEntry,
    CreateException,
    DL,
    DQuotSocket,
    DevlinkSocket,
    DiagSocket,
    Ethtool,
    EventSocket,
    GenericNetlinkSocket,
    IPBatch,
    IPDB,
    IPQSocket,
    IPRSocket,
    IPRoute,
    IPSet,
    IPVS,
    IPVSDest,
    IPVSService,
    IPVSSocket,
    IW,
    L2tp,
    MPTCP,
    NDB,
    NFCTSocket,
    NFTSocket,
    NL80211,
    NSPopen,
    NetNS,
    NetlinkDecodeError,
    NetlinkDumpInterrupted,
    NetlinkError,
    NlEthtool,
    Plan9ClientSocket,
    Plan9ServerSocket,
    ProcEventSocket,
    RawIPRoute,
    TaskStats,
    ThermalEventSocket,
    UeventSocket,
    WiSet,
    WireGuard,
    log,
    ss2,
]

__all__ = []
__all__.extend(modules)



import os
import socket
import subprocess
from socket import AF_INET, AF_INET6

from pyroute2.common import dqn2int
from pyroute2.netlink import NLM_F_DUMP, NLM_F_MULTI, NLM_F_REQUEST, NLMSG_DONE
from pyroute2.netlink.rtnl import (
    RTM_GETADDR,
    RTM_GETLINK,
    RTM_GETNEIGH,
    RTM_GETROUTE,
    RTM_NEWADDR,
    RTM_NEWLINK,
    RTM_NEWNEIGH,
    RTM_NEWROUTE,
)
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.netlink.rtnl.marshal import MarshalRtnl


class IPRoute(object):
    """
    macOS-specific IPRoute class (proof-of-concept).
    Attempts to fetch interfaces, addresses, neighbors, and routes
    by parsing macOS command output (ifconfig, netstat, arp, ndp, etc.).
    """

    def __init__(self, *argv, **kwarg):
        self.marshal = MarshalRtnl()
        self.target = kwarg.get('target') or 'localhost'
        self._outq = kwarg.get('_outq', None)
        self._pfdw = kwarg.get('_pfdw', None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def clone(self):
        return self

    def close(self, code=None):
        pass

    def bind(self, *argv, **kwarg):
        pass

    def getsockopt(self, *argv, **kwarg):
        return 1024 * 1024

    def sendto_gate(self, msg, addr):
        """
        Handle incoming netlink requests (simulation for macOS).
        """
        cmd = msg['header']['type']
        flags = msg['header']['flags']
        seq = msg['header']['sequence_number']

        # Only handle dump requests
        if flags != NLM_F_REQUEST | NLM_F_DUMP:
            return

        if cmd == RTM_GETLINK:
            rtype = RTM_NEWLINK
            ret = self.get_links()
        elif cmd == RTM_GETADDR:
            rtype = RTM_NEWADDR
            ret = self.get_addr()
        elif cmd == RTM_GETROUTE:
            rtype = RTM_NEWROUTE
            ret = self.get_routes()
        elif cmd == RTM_GETNEIGH:
            rtype = RTM_NEWNEIGH
            ret = self.get_neighbours()
        else:
            ret = []

        # Set response type and finalize
        for r in ret:
            r['header']['type'] = rtype
            r['header']['flags'] = NLM_F_MULTI
            r['header']['sequence_number'] = seq

        done_msg = type(msg)()
        done_msg['header']['type'] = NLMSG_DONE
        done_msg['header']['sequence_number'] = seq
        ret.append(done_msg)

        # If using an output queue (as done in Windows code),
        # write data there
        if self._outq and self._pfdw:
            data = b''
            for r in ret:
                r.encode()
                data += r.data
            self._outq.put(data)
            os.write(self._pfdw, b'\0')

    def _parse_ifconfig(self):
        """
        Parse `ifconfig` output to gather interface info, MAC, and addresses.
        Returns a dictionary:
        {
          'interfaces': [ifinfmsg, ifinfmsg, ...],
          'addresses':  [ifaddrmsg, ifaddrmsg, ...]
        }
        """
        ret = {'interfaces': [], 'addresses': []}

        try:
            output = subprocess.check_output(["ifconfig"]).decode(
                "utf-8", errors="replace"
            )
        except (subprocess.CalledProcessError, FileNotFoundError):
            return ret

        blocks = output.strip().split("\n\n")

        for block in blocks:
            lines = block.strip().splitlines()
            if not lines:
                continue

            # The first line usually has "en0: flags=..."
            first_line = lines[0]
            iface_name = first_line.split(":")[0].strip()

            # Attempt to fetch interface index and MAC address
            try:
                iface_index = socket.if_nametoindex(iface_name)
            except OSError:
                iface_index = 0  # fallback if something goes wrong

            mac_addr = "00:00:00:00:00:00"
            for line in lines:
                # Lines with "ether " on macOS contain MAC, e.g.:
                # "ether 00:1c:42:aa:bb:cc"
                if line.strip().startswith("ether "):
                    parts = line.strip().split()
                    if len(parts) >= 2:
                        mac_addr = parts[1]
                    break

            # Build ifinfmsg
            spec_if = {
                'index': iface_index,
                'attrs': (
                    ['IFLA_ADDRESS', mac_addr],
                    ['IFLA_IFNAME', iface_name],
                ),
            }
            msg_if = ifinfmsg().load(spec_if)
            msg_if['header']['target'] = self.target
            msg_if['header']['type'] = RTM_NEWLINK
            # remove 'value' if present
            msg_if.pop('value', None)
            ret['interfaces'].append(msg_if)

            # Look for IPv4 or IPv6 addresses
            # "inet 192.168.1.10 netmask 0xffffff00 broadcast 192.168.1.255"
            # "inet6 fe80::... prefixlen 64 ..."
            for line in lines[1:]:
                parts = line.strip().split()
                if line.strip().startswith("inet "):
                    # IPv4
                    try:
                        ip_idx = parts.index("inet") + 1
                        ip_addr = parts[ip_idx]
                        mask_str = "255.255.255.0"  # fallback
                        if "netmask" in parts:
                            mask_idx = parts.index("netmask") + 1
                            mask_hex = parts[mask_idx]
                            if mask_hex.startswith("0x"):
                                nm_value = int(mask_hex, 16)
                                dotted_mask = []
                                for _ in range(4):
                                    dotted_mask.insert(0, str(nm_value & 0xFF))
                                    nm_value >>= 8
                                mask_str = ".".join(dotted_mask)

                        prefix_len = dqn2int(mask_str)

                        spec_addr = {
                            'index': iface_index,
                            'family': AF_INET,
                            'prefixlen': prefix_len,
                            'attrs': (
                                ['IFA_ADDRESS', ip_addr],
                                ['IFA_LOCAL', ip_addr],
                                ['IFA_LABEL', iface_name],
                            ),
                        }
                        msg_addr = ifaddrmsg().load(spec_addr)
                        msg_addr['header']['target'] = self.target
                        msg_addr['header']['type'] = RTM_NEWADDR
                        msg_addr.pop('value', None)
                        ret['addresses'].append(msg_addr)
                    except (ValueError, IndexError):
                        continue
                elif line.strip().startswith("inet6 "):
                    # IPv6
                    try:
                        ip_idx = parts.index("inet6") + 1
                        ip_addr = parts[ip_idx]
                        prefix_len = 64  # default if not found

                        # On macOS you'll often see
                        # "inet6 fe80::xxxx prefixlen 64 ..."
                        if "prefixlen" in parts:
                            pre_idx = parts.index("prefixlen") + 1
                            prefix_len = int(parts[pre_idx])

                        spec_addr = {
                            'index': iface_index,
                            'family': AF_INET6,
                            'prefixlen': prefix_len,
                            'attrs': (
                                ['IFA_ADDRESS', ip_addr],
                                ['IFA_LOCAL', ip_addr],
                                ['IFA_LABEL', iface_name],
                            ),
                        }
                        msg_addr = ifaddrmsg().load(spec_addr)
                        msg_addr['header']['target'] = self.target
                        msg_addr['header']['type'] = RTM_NEWADDR
                        msg_addr.pop('value', None)
                        ret['addresses'].append(msg_addr)
                    except (ValueError, IndexError):
                        continue

        return ret

    def _parse_routes(self):
        """
        Parse `netstat -rn` to retrieve routing table entries.
        Return list of ifinfmsg-like route objects, though strictly
        they'd be `rtmsg` in a real Netlink environment.
        For demonstration, we just return RTM_NEWROUTE messages
        with minimal info: destination, gateway, interface index, etc.
        """
        routes = []
        try:
            output = subprocess.check_output(["netstat", "-rn"]).decode(
                "utf-8", errors="replace"
            )
        except (subprocess.CalledProcessError, FileNotFoundError):
            return routes

        # On macOS, `netstat -rn` has headers like:
        # Destination        Gateway            Flags        Netif Expire
        # default            192.168.1.1        UGSc         en0
        # 127                127.0.0.1          UCS          lo0
        # ...
        lines = output.strip().splitlines()
        # skip header lines
        # find the header first
        start_index = 0
        for i, line in enumerate(lines):
            if "Destination" in line and "Gateway" in line:
                start_index = i + 1
                break

        # parse from start_index onward
        for line in lines[start_index:]:
            parts = line.split()
            if len(parts) < 4:
                continue
            destination, gateway, flags, netif = parts[:4]
            # Attempt to get interface index
            try:
                idx = socket.if_nametoindex(netif)
            except OSError:
                idx = 0

            # We'll store the route as ifinfmsg for
            # demonstration, but real netlink
            # code would use `rtmsg`. We'll keep it
            # consistent with the rest of this
            # POC approach.
            spec_route = {
                'index': idx,
                # not strictly correct, but placing
                # them in attrs for demonstration
                'attrs': (
                    ['ROUTE_DST', destination],
                    ['ROUTE_GATEWAY', gateway],
                    ['ROUTE_FLAGS', flags],
                    ['ROUTE_IFNAME', netif],
                ),
            }
            # We'll build it as ifinfmsg with a 'type' = RTM_NEWROUTE
            msg_route = ifinfmsg().load(spec_route)
            msg_route['header']['target'] = self.target
            msg_route['header']['type'] = RTM_NEWROUTE
            msg_route.pop('value', None)
            routes.append(msg_route)

        return routes

    def _parse_arp_neighbors(self):
        """
        Parse `arp -an` for IPv4 neighbors and (optionally)
        `ndp -an` for IPv6.
        Returns a list of ifinfmsg (or ifaddrmsg)
        objects simulating neighbor entries.
        """
        neighbors = []
        # IPv4 neighbors (ARP)
        try:
            output = subprocess.check_output(["arp", "-an"]).decode(
                "utf-8", errors="replace"
            )
            # Lines look like:
            # "? (192.168.1.10) at 00:1c:42:xx:yy:zz on en0 ifscope [ethernet]"
            for line in output.strip().splitlines():
                line = line.strip()
                if not line:
                    continue
                # Quick parse:
                # 1) IP in parentheses
                # 2) MAC after "at"
                # 3) interface name after "on"
                parts = line.split()
                # example parts:
                # ["?", "(192.168.1.10)",
                # "at", "00:1c:42:xx:yy:zz",
                # "on", "en0", ...]
                if len(parts) < 7:
                    continue
                ip_str = parts[1].strip("()")
                mac_str = parts[3]
                iface_str = parts[5]
                try:
                    idx = socket.if_nametoindex(iface_str)
                except OSError:
                    idx = 0

                spec_neigh = {
                    'index': idx,
                    'attrs': (
                        ['NEIGH_IP', ip_str],
                        ['NEIGH_LLADDR', mac_str],
                        ['NEIGH_IFNAME', iface_str],
                    ),
                }
                msg_neigh = ifinfmsg().load(spec_neigh)
                msg_neigh['header']['target'] = self.target
                msg_neigh['header']['type'] = RTM_NEWNEIGH
                msg_neigh.pop('value', None)
                neighbors.append(msg_neigh)
        except (subprocess.CalledProcessError, FileNotFoundError):
            pass

        # Optional: IPv6 neighbors (NDP)
        try:
            output = subprocess.check_output(["ndp", "-an"]).decode(
                "utf-8", errors="replace"
            )
            # Lines look like:
            # "fe80::1%lo0            lladdr 00:00:00:...  router  STALE"
            for line in output.strip().splitlines():
                line = line.strip()
                if not line:
                    continue
                parts = line.split()
                if len(parts) < 4:
                    continue
                # typical parts:
                # ["fe80::1%lo0",
                # "lladdr", "00:00:00:..",
                # "router", "STALE"]
                # or ["2001:db8::1234%en0",
                # "lladdr",
                # "00:11:22:33:44:55",
                # "REACHABLE"]
                addr_part = parts[0]
                # separate out the interface from the IP
                if "%" in addr_part:
                    ip_str, iface_str = addr_part.split("%", 1)
                else:
                    ip_str = addr_part
                    iface_str = "??"
                mac_str = None
                if "lladdr" in parts:
                    ll_idx = parts.index("lladdr") + 1
                    if ll_idx < len(parts):
                        mac_str = parts[ll_idx]
                if not mac_str:
                    continue
                try:
                    idx = socket.if_nametoindex(iface_str)
                except OSError:
                    idx = 0

                spec_neigh = {
                    'index': idx,
                    'attrs': (
                        ['NEIGH_IP', ip_str],
                        ['NEIGH_LLADDR', mac_str],
                        ['NEIGH_IFNAME', iface_str],
                    ),
                }
                msg_neigh = ifinfmsg().load(spec_neigh)
                msg_neigh['header']['target'] = self.target
                msg_neigh['header']['type'] = RTM_NEWNEIGH
                msg_neigh.pop('value', None)
                neighbors.append(msg_neigh)
        except (subprocess.CalledProcessError, FileNotFoundError):
            pass

        return neighbors

    def dump(self, groups=None):
        # Return all info in sequence
        for method in (
            self.get_links,
            self.get_addr,
            self.get_neighbours,
            self.get_routes,
        ):
            for msg in method():
                yield msg

    def get_links(self, *argv, **kwarg):
        '''
        {
          'attrs': (['IFLA_ADDRESS', '9a:9d:81:90:d2:8f'],
                    ['IFLA_IFNAME', 'lo0']),
          'header': {'target': 'localhost', 'type': 16},
          'index': 1
        }
        '''
        return self._parse_ifconfig()['interfaces']

    def get_addr(self, *argv, **kwarg):
        '''
        {
          'attrs': (['IFA_ADDRESS', '127.0.0.1'],
                    ['IFA_LOCAL', '127.0.0.1'],
                    ['IFA_LABEL', 'lo0']),
          'header': {'target': 'localhost', 'type': 20},
          'index': 1,
          'family': <AddressFamily.AF_INET: 2>,
          'prefixlen': 8
        }
        '''
        return self._parse_ifconfig()['addresses']

    def get_neighbours(self, *argv, **kwarg):
        return self._parse_arp_neighbors()

    def get_routes(self, *argv, **kwarg):
        return self._parse_routes()


class RawIPRoute(IPRoute):
    pass


class ChaoticIPRoute:
    """
    Placeholder, mirroring its Windows counterpart.
    """

    def __init__(self, *argv, **kwarg):
        raise NotImplementedError()


class NetNS:
    """
    Another placeholder class to match the Windows code structure.
    """

    def __init__(self, *argv, **kwarg):
        raise NotImplementedError()



'''Timers to manage lease rebinding, renewal & expiration.'''

import asyncio
import dataclasses
from logging import getLogger
from typing import Optional

from pyroute2.dhcp.fsm import AsyncCallback
from pyroute2.dhcp.leases import Lease

LOG = getLogger(__name__)


@dataclasses.dataclass
class LeaseTimers:
    '''Manage callbacks associated with DHCP leases.'''

    renewal: Optional[asyncio.TimerHandle] = None
    rebinding: Optional[asyncio.TimerHandle] = None
    expiration: Optional[asyncio.TimerHandle] = None

    def cancel(self) -> None:
        '''Cancel all current timers.'''
        for timer_name in ('renewal', 'rebinding', 'expiration'):
            self._reset_timer(timer_name)

    def _reset_timer(self, timer_name: str) -> None:
        '''Cancel a timer and set it to None.'''
        if timer_name not in ('renewal', 'rebinding', 'expiration'):
            raise ValueError(f'Unknown timer {timer_name!r}')
        if timer := getattr(self, timer_name):
            assert isinstance(timer, asyncio.TimerHandle)
            if not timer.cancelled():
                # FIXME: how do we know a timer wasn't cancelled ?
                # this causes spurious logs
                LOG.debug('Canceling %s timer', timer_name)
                timer.cancel()
            setattr(self, timer_name, None)

    def arm(self, lease: Lease, **callbacks: AsyncCallback) -> None:
        '''Reset & arm timers from a `Lease`.

        `callbacks` must be async callables with no arguments
        that will be called when the associated timer expires.
        '''
        self.cancel()
        loop = asyncio.get_running_loop()

        for timer_name, async_callback in callbacks.items():
            self._reset_timer(timer_name)
            lease_time = getattr(lease, f'{timer_name}_in')
            if not lease_time:
                LOG.debug('%s time is infinite', timer_name)
                continue
            if lease_time < 0.0:
                LOG.debug('%s is in the past', timer_name)
                continue
            LOG.info('Scheduling %s in %.2fs', timer_name, lease_time)
            timer = loop.call_later(
                lease_time, self._create_timer_task, timer_name, async_callback
            )
            setattr(self, timer_name, timer)

    def _create_timer_task(
        self, timer_name: str, async_callback: AsyncCallback
    ) -> None:
        ''' 'Internal callback for loop.call_later.

        Creates a Task that runs the async_callback.
        '''
        asyncio.create_task(
            async_callback(), name=f"{timer_name} timer callback"
        )
